#!BPY

"""
Name: 'MD2 -new- (.md2)...'
Blender: 245
Group: 'Export'
Tooltip: 'Export to Quake2 file format (.md2) -new-'
"""

__version__ = '0.1'
__author__  = [ 'Damien Thébault', 'Erwan Mathieu' ]
__url__     = [ 'https://metabolsim.no-ip.org/trac/metabolsim' ]
__email__   = [ 'Damien Thébault <damien!thebault#laposte!net>', 'Erwan Mathieu <cyberwan#laposte!net>' ]

__bpydoc__ = """\
This script exports a mesh object to the Quake2 "MD2" file format.

If you want to choose the name of the animations, you have to go to the timeline, choose the starting frame, choose 'Frame', 'Add Marker' (M) and then set a name with 'Frame', 'Name Marker' (Ctrl M).

The frames are named <frameName><N> with :<br>
 - <N> the frame number<br>
 - <frameName> the name choosen at the last marker (or 'frame' if the last marker has no name or if there is no last marker)

Skins are set using image textures in materials. If a skin path is too long (>63 chars), the basename will be used.

Usually, importers expect that the name don't have any digit (to be able to detect where the frame name starts and where the frame number starts.

Thanks to Bob Holcomb for MD2_NORMALS taken from his exporter.<br>
Thanks to David Henry for the documentation about the MD2 file format.
"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****


import Blender
import struct
import random

# TODO : if mode is 'solid', separate vertices and use the face's normal
# TODO : save skins

MD2_NORMALS=((-0.525731, 0.000000, 0.850651),
             (-0.442863, 0.238856, 0.864188),
             (-0.295242, 0.000000, 0.955423),
             (-0.309017, 0.500000, 0.809017),
             (-0.162460, 0.262866, 0.951056),
             ( 0.000000, 0.000000, 1.000000),
             ( 0.000000, 0.850651, 0.525731),
             (-0.147621, 0.716567, 0.681718),
             ( 0.147621, 0.716567, 0.681718),
             ( 0.000000, 0.525731, 0.850651),
             ( 0.309017, 0.500000, 0.809017),
             ( 0.525731, 0.000000, 0.850651),
             ( 0.295242, 0.000000, 0.955423),
             ( 0.442863, 0.238856, 0.864188),
             ( 0.162460, 0.262866, 0.951056),
             (-0.681718, 0.147621, 0.716567),
             (-0.809017, 0.309017, 0.500000),
             (-0.587785, 0.425325, 0.688191),
             (-0.850651, 0.525731, 0.000000),
             (-0.864188, 0.442863, 0.238856),
             (-0.716567, 0.681718, 0.147621),
             (-0.688191, 0.587785, 0.425325),
             (-0.500000, 0.809017, 0.309017),
             (-0.238856, 0.864188, 0.442863),
             (-0.425325, 0.688191, 0.587785),
             (-0.716567, 0.681718,-0.147621),
             (-0.500000, 0.809017,-0.309017),
             (-0.525731, 0.850651, 0.000000),
             ( 0.000000, 0.850651,-0.525731),
             (-0.238856, 0.864188,-0.442863),
             ( 0.000000, 0.955423,-0.295242),
             (-0.262866, 0.951056,-0.162460),
             ( 0.000000, 1.000000, 0.000000),
             ( 0.000000, 0.955423, 0.295242),
             (-0.262866, 0.951056, 0.162460),
             ( 0.238856, 0.864188, 0.442863),
             ( 0.262866, 0.951056, 0.162460),
             ( 0.500000, 0.809017, 0.309017),
             ( 0.238856, 0.864188,-0.442863),
             ( 0.262866, 0.951056,-0.162460),
             ( 0.500000, 0.809017,-0.309017),
             ( 0.850651, 0.525731, 0.000000),
             ( 0.716567, 0.681718, 0.147621),
             ( 0.716567, 0.681718,-0.147621),
             ( 0.525731, 0.850651, 0.000000),
             ( 0.425325, 0.688191, 0.587785),
             ( 0.864188, 0.442863, 0.238856),
             ( 0.688191, 0.587785, 0.425325),
             ( 0.809017, 0.309017, 0.500000),
             ( 0.681718, 0.147621, 0.716567),
             ( 0.587785, 0.425325, 0.688191),
             ( 0.955423, 0.295242, 0.000000),
             ( 1.000000, 0.000000, 0.000000),
             ( 0.951056, 0.162460, 0.262866),
             ( 0.850651,-0.525731, 0.000000),
             ( 0.955423,-0.295242, 0.000000),
             ( 0.864188,-0.442863, 0.238856),
             ( 0.951056,-0.162460, 0.262866),
             ( 0.809017,-0.309017, 0.500000),
             ( 0.681718,-0.147621, 0.716567),
             ( 0.850651, 0.000000, 0.525731),
             ( 0.864188, 0.442863,-0.238856),
             ( 0.809017, 0.309017,-0.500000),
             ( 0.951056, 0.162460,-0.262866),
             ( 0.525731, 0.000000,-0.850651),
             ( 0.681718, 0.147621,-0.716567),
             ( 0.681718,-0.147621,-0.716567),
             ( 0.850651, 0.000000,-0.525731),
             ( 0.809017,-0.309017,-0.500000),
             ( 0.864188,-0.442863,-0.238856),
             ( 0.951056,-0.162460,-0.262866),
             ( 0.147621, 0.716567,-0.681718),
             ( 0.309017, 0.500000,-0.809017),
             ( 0.425325, 0.688191,-0.587785),
             ( 0.442863, 0.238856,-0.864188),
             ( 0.587785, 0.425325,-0.688191),
             ( 0.688191, 0.587785,-0.425325),
             (-0.147621, 0.716567,-0.681718),
             (-0.309017, 0.500000,-0.809017),
             ( 0.000000, 0.525731,-0.850651),
             (-0.525731, 0.000000,-0.850651),
             (-0.442863, 0.238856,-0.864188),
             (-0.295242, 0.000000,-0.955423),
             (-0.162460, 0.262866,-0.951056),
             ( 0.000000, 0.000000,-1.000000),
             ( 0.295242, 0.000000,-0.955423),
             ( 0.162460, 0.262866,-0.951056),
             (-0.442863,-0.238856,-0.864188),
             (-0.309017,-0.500000,-0.809017),
             (-0.162460,-0.262866,-0.951056),
             ( 0.000000,-0.850651,-0.525731),
             (-0.147621,-0.716567,-0.681718),
             ( 0.147621,-0.716567,-0.681718),
             ( 0.000000,-0.525731,-0.850651),
             ( 0.309017,-0.500000,-0.809017),
             ( 0.442863,-0.238856,-0.864188),
             ( 0.162460,-0.262866,-0.951056),
             ( 0.238856,-0.864188,-0.442863),
             ( 0.500000,-0.809017,-0.309017),
             ( 0.425325,-0.688191,-0.587785),
             ( 0.716567,-0.681718,-0.147621),
             ( 0.688191,-0.587785,-0.425325),
             ( 0.587785,-0.425325,-0.688191),
             ( 0.000000,-0.955423,-0.295242),
             ( 0.000000,-1.000000, 0.000000),
             ( 0.262866,-0.951056,-0.162460),
             ( 0.000000,-0.850651, 0.525731),
             ( 0.000000,-0.955423, 0.295242),
             ( 0.238856,-0.864188, 0.442863),
             ( 0.262866,-0.951056, 0.162460),
             ( 0.500000,-0.809017, 0.309017),
             ( 0.716567,-0.681718, 0.147621),
             ( 0.525731,-0.850651, 0.000000),
             (-0.238856,-0.864188,-0.442863),
             (-0.500000,-0.809017,-0.309017),
             (-0.262866,-0.951056,-0.162460),
             (-0.850651,-0.525731, 0.000000),
             (-0.716567,-0.681718,-0.147621),
             (-0.716567,-0.681718, 0.147621),
             (-0.525731,-0.850651, 0.000000),
             (-0.500000,-0.809017, 0.309017),
             (-0.238856,-0.864188, 0.442863),
             (-0.262866,-0.951056, 0.162460),
             (-0.864188,-0.442863, 0.238856),
             (-0.809017,-0.309017, 0.500000),
             (-0.688191,-0.587785, 0.425325),
             (-0.681718,-0.147621, 0.716567),
             (-0.442863,-0.238856, 0.864188),
             (-0.587785,-0.425325, 0.688191),
             (-0.309017,-0.500000, 0.809017),
             (-0.147621,-0.716567, 0.681718),
             (-0.425325,-0.688191, 0.587785),
             (-0.162460,-0.262866, 0.951056),
             ( 0.442863,-0.238856, 0.864188),
             ( 0.162460,-0.262866, 0.951056),
             ( 0.309017,-0.500000, 0.809017),
             ( 0.147621,-0.716567, 0.681718),
             ( 0.000000,-0.525731, 0.850651),
             ( 0.425325,-0.688191, 0.587785),
             ( 0.587785,-0.425325, 0.688191),
             ( 0.688191,-0.587785, 0.425325),
             (-0.955423, 0.295242, 0.000000),
             (-0.951056, 0.162460, 0.262866),
             (-1.000000, 0.000000, 0.000000),
             (-0.850651, 0.000000, 0.525731),
             (-0.955423,-0.295242, 0.000000),
             (-0.951056,-0.162460, 0.262866),
             (-0.864188, 0.442863,-0.238856),
             (-0.951056, 0.162460,-0.262866),
             (-0.809017, 0.309017,-0.500000),
             (-0.864188,-0.442863,-0.238856),
             (-0.951056,-0.162460,-0.262866),
             (-0.809017,-0.309017,-0.500000),
             (-0.681718, 0.147621,-0.716567),
             (-0.681718,-0.147621,-0.716567),
             (-0.850651, 0.000000,-0.525731),
             (-0.688191, 0.587785,-0.425325),
             (-0.587785, 0.425325,-0.688191),
             (-0.425325, 0.688191,-0.587785),
             (-0.425325,-0.688191,-0.587785),
             (-0.587785,-0.425325,-0.688191),
             (-0.688191,-0.587785,-0.425325))

class MD2:
	def __init__(self, anim, basename):
		self.anim = anim
		self.basename = basename
		self.object = None
		return

	def setObject(self, object):
		self.object = object

	def write(self, filename):
		self.ident = 'IDP2'
		self.version = 8

		self.skinwidth = 2**10-1
		self.skinheight = 2**10-1

		# self.framesize : see below

		mesh = self.object.getData(mesh=True)

		skins = Util.getSkins(mesh)

		self.num_skins = len(skins)
		self.num_xyz = len(mesh.verts)
		self.num_st = len(mesh.faces)*3
		self.num_tris = len(mesh.faces)
		self.num_glcmds = self.num_tris * (1+3*3) + 1
		if self.anim:
			self.num_frames = 1 + Blender.Get('endframe') \
			                    - Blender.Get('staframe')
		else:
			self.num_frames = 1

		self.framesize = 40+4*self.num_xyz

		self.ofs_skins = 68 # size of the header
		self.ofs_st = self.ofs_skins + 64*self.num_skins
		self.ofs_tris = self.ofs_st + 4*self.num_st
		self.ofs_frames = self.ofs_tris + 12*self.num_tris
		self.ofs_glcmds = self.ofs_frames + self.framesize*self.num_frames
		self.ofs_end = self.ofs_glcmds + 4*self.num_glcmds

		file = open(filename, 'wb')
		try:
			bin = struct.pack('<4s16i',
			                  self.ident,
			                  self.version,
			                  self.skinwidth,
			                  self.skinheight,
			                  self.framesize,
			                  self.num_skins,
			                  self.num_xyz,
			                  self.num_st,
			                  self.num_tris,
			                  self.num_glcmds,
			                  self.num_frames,
			                  self.ofs_skins,
			                  self.ofs_st,
			                  self.ofs_tris,
			                  self.ofs_frames,
			                  self.ofs_glcmds,
			                  self.ofs_end)
			file.write(bin) # header

			for skin in skins:
				if len(skin) > 63 or self.basename:
					skin = Blender.sys.basename(skin)
				bin = struct.pack('<64s', skin[0:63])
				file.write(bin) # skin name

			for face in mesh.faces:
				try:
					uvs = face.uv
				except:
					uvs = ([0,0],[0,0],[0,0])

				# (u,v) in blender -> (u,1-v)
				bin = struct.pack('<6h',
				                  uvs[0][0]*self.skinwidth,
				                  (1-uvs[0][1])*self.skinheight,
				                  uvs[1][0]*self.skinwidth,
				                  (1-uvs[1][1])*self.skinheight,
				                  uvs[2][0]*self.skinwidth,
				                  (1-uvs[2][1])*self.skinheight,
				                  )
				file.write(bin) # uv
				# (uv index is : face.index*3+i)

			for face in mesh.faces:
				# 0,2,1 for good cw/ccw
				bin = struct.pack('<3h',
				                  face.verts[0].index,
				                  face.verts[2].index,
				                  face.verts[1].index,
				                )
				file.write(bin) # vert index
				bin = struct.pack('<3h',
				                  face.index*3 + 0,
				                  face.index*3 + 2,
				                  face.index*3 + 1,
				                  )
				file.write(bin) # uv index

			if self.anim:
				min = None
				max = None
				for frame in range(1, self.num_frames+1):
					Blender.Set('curframe', frame)
					(min, max) = self.findMinMax(min, max)

				timeLine = Blender.Scene.GetCurrent().getTimeLine()
				frameNames = timeLine.getMarked()
				name = 'frame'
				for frame in range(1, self.num_frames+1):
					percent = (frame-Blender.Get('staframe')) / \
					          ( 1. + self.num_frames)
					Blender.Window.DrawProgressBar(percent,
					           str(int(percent*100)) + '%' + \
					           ' : frame ' + str(frame))
					Blender.Set('curframe', frame)
					if frame in frameNames:
						name = frameNames[frame][0]
						if name == '':
							name = 'frame'
					self.outFrame(file, min, max, name + str(frame))
			else:
				(min, max) = self.findMinMax()
				self.outFrame(file, min, max)

			# gl commands
			for face in mesh.faces:
				try:
					uvs = face.uv
				except:
					uvs = ([0,0],[0,0],[0,0])
				bin = struct.pack('<i', 3)
				file.write(bin)
				# 0,2,1 for good cw/ccw (also flips/inverts normal)
				for vert in [0,2,1]:
					# (u,v) in blender -> (u,1-v)
					bin = struct.pack('<ffI',
						uvs[vert][0],
						(1.0 - uvs[vert][1]),
						face.verts[vert].index)
					file.write(bin)
			# NULL command
			bin = struct.pack('<I', 0)
			file.write(bin)
		finally:
			file.close()

	def findMinMax(self, min=None, max=None):
		mesh = Blender.Mesh.New()
		mesh.getFromObject(self.object.name)
		mesh.transform(self.object.getMatrix())
		mesh.transform(Blender.Mathutils.RotationMatrix(90, 4, 'z')) # Hoehrer: rotate 90 degrees

		if min == None:
			min = [mesh.verts[0].co[0],
			       mesh.verts[0].co[1],
			       mesh.verts[0].co[2]]
		if max == None:
			max = [mesh.verts[0].co[0],
			       mesh.verts[0].co[1],
			       mesh.verts[0].co[2]]

		for vert in mesh.verts:
			for i in range(3):
				if vert.co[i] < min[i]:
					min[i] = vert.co[i]
				if vert.co[i] > max[i]:
					max[i] = vert.co[i]

		return (min, max)

	def outFrame(self, file, min, max, frameName = 'frame'):
		mesh = Blender.Mesh.New()
		mesh.getFromObject(self.object.name)
		mesh.transform(self.object.getMatrix())
		mesh.transform(Blender.Mathutils.RotationMatrix(90, 4, 'z')) # Hoehrer: rotate 90 degrees

		bin = struct.pack('<6f16s',
		                  self.object.getSize()[0]*(max[0]-min[0])/255.,
		                  self.object.getSize()[1]*(max[1]-min[1])/255.,
		                  self.object.getSize()[2]*(max[2]-min[2])/255.,
		                  min[0],
		                  min[1],
		                  min[2],
		                  frameName)
		file.write(bin) # frame header
		for vert in mesh.verts:
			for i in range(162):
				dot =  vert.no[1]*MD2_NORMALS[i][0] + \
				      -vert.no[0]*MD2_NORMALS[i][1] + \
				       vert.no[2]*MD2_NORMALS[i][2]
				if (i==0) or (dot > maxDot):
					maxDot = dot
					bestNormalIndex = i

			bin = struct.pack('<4B',
			                  int((vert.co[0]-min[0])/(max[0]-min[0])*255.),
			                  int((vert.co[1]-min[1])/(max[1]-min[1])*255.),
			                  int((vert.co[2]-min[2])/(max[2]-min[2])*255.),
			                  bestNormalIndex)
			file.write(bin) # vertex

class Util:
	@staticmethod
	def pickName():
		name = '_MD2Obj_'+str(random.random())
		return name[0:20]

	@staticmethod
	def duplicateObject(object, name):
		object.select(True)
		# duplicate selected object
		Blender.Object.Duplicate(mesh=True)
		# select the duplicated object
		object = Blender.Object.GetSelected()[0]
		object.setName(name)
		return object

	@staticmethod
	def getTriangMesh(object):
		mesh = object.getData(mesh=True)
		mesh.sel = True
		mesh.quadToTriangle()
		return mesh

	@staticmethod
	def getSkins(mesh):
		skins = []
		for material in mesh.materials:
			for texture in material.getTextures():
				if texture != None and texture.tex != None:
					if texture.tex.getType() == "Image":
						imgfile = texture.tex.getImage().getFilename()[2:]
						skins.append(imgfile)
		return skins

class ObjectInfo:
	def __init__(self, object):
		self.triang = False
		self.verts = -1
		self.faces = -1
		self.status = ('','')

		self.ismesh = object and object.getType() == 'Mesh'

		if self.ismesh:
			originalObject = object

			mesh = object.getData(mesh=True)

			self.skins = Util.getSkins(mesh)

			for face in mesh.faces:
				if len(face) == 4:
					self.triang = True
					break

			tmpObjectName = Util.pickName()
			try:
				if self.triang:
					object = Util.duplicateObject(object, tmpObjectName)
					mesh = Util.getTriangMesh(object)

				self.status = (str(len(mesh.verts)) + ' vertices',
							str(len(mesh.faces)) + ' faces')
			finally:
				if object.getName() == tmpObjectName:
					originalObject.select(True)
					Blender.Scene.GetCurrent().objects.unlink(object)

class GUI:
	id_export   = 1
	id_cancel   = 2
	id_anim     = 3
	id_update   = 4
	id_help     = 5
	id_basename = 6
	val_anim     = True
	val_basename = True

	str_export = 'Export to MD2...'

	def __init__(self):
		try:
			self.object = Blender.Object.GetSelected()[0]
		except:
			self.object = None

		self.info = ObjectInfo(self.object)

	def draw(self):
		(width, height) = Blender.Window.GetAreaSize()

		# export - cancel - help
		eWidth = Blender.Draw.GetStringWidth(self.str_export)+20
		if self.info.ismesh:
			Blender.Draw.PushButton(self.str_export, self.id_export,
			                        width-eWidth-11, height-37, eWidth, 21)

		Blender.Draw.PushButton('Cancel', self.id_cancel,
		                        width-eWidth-11, height-37-24, eWidth, 21)

		Blender.Draw.PushButton('Help', self.id_help,
		                        width-eWidth-11, height-100, eWidth, 21)

		# *n* frames - export animation
		left    = width-eWidth-140
		farLeft = left-120
		nframes = 1+Blender.Get('endframe')-Blender.Get('staframe')
		Blender.Draw.Label(str(nframes)+' frames',
		                   farLeft, height-100, left-40-farLeft, 20)
		if nframes <= 512:
			Blender.Draw.Toggle('Export animation',
			                    self.id_anim, left-40, height-100, 140, 20,
			                    self.val_anim)
		else:
			self.val_anim = False
			Blender.Draw.Label('Too many frames!',
			                   left, height-100, 100, 20)
			Blender.Draw.Label('(' + str(nframes) + ', max : 512)',
			                   left, height-115, 100, 20)
		# *n* skins - use full path
		skinstatus = str(len(self.info.skins))+' skin'
		if len(self.info.skins) > 1:
			skinstatus += 's'
		Blender.Draw.Label(skinstatus,
		                   farLeft, height-120, left-40-farLeft, 20)
		Blender.Draw.Toggle('Export only basenames',
		                    self.id_basename, left-40, height-120, 140, 20,
		                    self.val_basename)
		# object : *name* - update
		if self.object != None:
			Blender.Draw.Label('Object : ' + self.object.getName(),
			                   farLeft, height-150, left-farLeft, 20)
		else:
			Blender.Draw.Label('Object : None',
			                   farLeft, height-150, left-farLeft, 20)
		if self.info.ismesh:
			if self.info.triang:
				status = '- Will be triangulated'
			else:
				status = '- No need to triangulate'
			Blender.Draw.Label(status,
			                   left, height-150, width-left, 20)

			for i in range(len(self.info.status)):
				Blender.Draw.Label('- ' + self.info.status[i],
				                   left, height-150-(i+1)*15,
				                   width-left, 20)

		else:
			status = 'Not a mesh!'
			Blender.Draw.Label(status,
			                   width-Blender.Draw.GetStringWidth(status)-40,
			                   height-37, eWidth, 21)

	def button_event(self, evt):
		if evt == self.id_export:
			Blender.Window.FileSelector(self.export, self.str_export,
			                            Blender.sys.makename(ext='.md2'))
		if evt == self.id_anim:
			self.val_anim = not self.val_anim
		if evt == self.id_basename:
			self.val_basename = not self.val_basename
		if evt == self.id_cancel:
			Blender.Draw.Exit()
		if evt == self.id_update:
			updateGuiObject()
			Blender.Draw.Redraw()
		if evt == self.id_help:
			Blender.ShowHelp('export_md2_new.py')

	def export(self, filename):

		if Blender.sys.exists(filename) == 1:
			overwrite = Blender.Draw.PupMenu('File already exists, overwrite?%t|Yes%x1|No%x0')
			if overwrite == 0:
				return

		Blender.Window.WaitCursor(True)

		object = self.object
		originalObject = object

		if object.getType() != 'Mesh':
			raise NameError, 'Selected object must be a mesh!'

		# different name each time or we can't unlink it later
		tmpObjectName = Util.pickName()

		mesh = object.getData(mesh=True)

		if self.info.triang:
			object = Util.duplicateObject(object, tmpObjectName)
			mesh = Util.getTriangMesh(object)

		if self.val_anim:
			frame = Blender.Get('curframe')

		try:
			md2 = MD2(self.val_anim, self.val_basename)
			md2.setObject(object)
			md2.write(filename)
		finally:
			if object.getName() == tmpObjectName:
				originalObject.select(True)
				Blender.Scene.GetCurrent().objects.unlink(object)
			if self.val_anim:
				Blender.Set('curframe', frame)
			Blender.Window.DrawProgressBar(1.0, '')
			Blender.Window.WaitCursor(False)
			Blender.Draw.Exit()


if __name__ == '__main__':
	gui = GUI()
	Blender.Draw.Register(gui.draw, None, gui.button_event)



#!BPY

# Blender exporter for UFO:AI.

# Copyright 2008 (c) Wrwrwr <http://www.wrwrwr.org>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# --------------------------------------------------------------------------
# Changelog
# 	0.2.1	mattn fixed missing texture and removed stepon
# 	0.2	WRWRWR Initial version
# --------------------------------------------------------------------------


"""
Name: 'UFO:AI (.map)'
Blender: 245
Group: 'Export'
Tooltip: 'Export to UFO:AI map format.'
"""

__author__ = 'Wrwrwr'
__version__ = '0.2.1'
__email__ = 'ufoai@wrwrwr.org'
__bpydoc__ = '''\
Exports current scene as an UFO:AI map.

See the full documentation at: http://www.wrwrwr.org/b2ufo/.

Simple convex meshes, lights, models and generic entities are exported.
'''

from math import *
from numpy import *
from numpy.linalg import *

from copy import copy, deepcopy
from cStringIO import StringIO
from datetime import datetime
from logging import Filter, basicConfig, debug, error, getLogger, info, warning
from os.path import exists, isabs, join, normpath, realpath, splitext
from os import sep
from sys import stderr
from time import clock

from Blender import *
import BPyMesh # for faceAngles



# planes for Q2 texture mapping; normal, its non zero indices
# "natural" texture mapping -- horrible, moreover the planes have different handedness
PLANES = (
(( 0,  0,  1), (0, 1)),			# floor, lh
(( 0,  0, -1), (0, 1)),			# ceiling, rh
(( 1,  0,  0), (1, 2)),			# west wall, lh
((-1,  0,  0), (1, 2)),			# east wall, rh
(( 0,  1,  0), (0, 2)),			# south wall, rh
(( 0, -1,  0), (0, 2))			# north wall, lh
)

# dummy object needed only for triangulation in splitting *** move to export
do = Object.New('Mesh', 'Dummy Exporter Object')

# dummy mesh used for exporting meshes without changing them *** move to export
dm = Mesh.New()

# ~rotation matrices used in texture mapping, *** move this to export (for configurable sum weighting)
a0 = matrix([[ 0,  1], [ 1,  0]])
a1 = matrix([[ 1,  0], [ 0, -1]])
asit = (a0 + a1).I.T



def normalize(vector):
	'''
	Normalize a numpy array. Should be a numpy function.
	'''
	return vector / norm(vector)



def needsSplitting(mesh, scale):
	'''
	Checks if it's ok to export mesh as it is. Must be strictly convex.
	Connectivity: checks if every edge is used by exactly two faces.
	Convexity: checks if every two normals of neighbouring faces intersect behind them.
	This should also work for all cases that are not polyhedra, but are exportable.
	Complexity: O(number of edges).
	*** check if the mesh is at all connected?
	*** are these tests enough
	'''
	# make a dictionary, where the edge is the key, and a list of faces that use it is the value
	esfs = dict([(e.key, []) for e in mesh.edges])
	for f in mesh.faces:
		for ek in f.edge_keys:
			esfs[ek].append(f)

	# check normals of every two neighbouring faces
	for ek, fs in esfs.iteritems():
		if len(fs) != 2:
			warning('\t\tNeeds splitting: An edge is not used by exactly 2 faces.')
			return True
		if dot(fs[0].no, fs[1].cent - fs[0].cent) >= 0:
			warning('\t\tNeeds splitting: Concave around: %s.' % around((fs[0].cent + fs[1].cent) / 2 / scale, 2))
			'''
			# sometimes a handy way to help in fixing those 'where is it concave?' things
			mesh.mode = Mesh.SelectModes['FACE']
			fs[0].sel = fs[1].sel = 1
			mesh = mesh.__copy__()
			sob = Object.New('Mesh', 'Concave Debug')
			Scene.getCurrent().objects.link(sob)
			sob.link(mesh)
			'''
			return True

	'''
	# check face areas, that's only needed with early coordinates rounding
	# such checks would only be needed without enlarging coordinates
	for f in mesh.faces:
		if f.area < minFaceArea:
			warning('\t\tNeeds splitting: A very small face present.')
			return True
	'''
	return False


# *** advanced, efficient splitting needed (e.g.: add as many faces as will allow to close polyhedron with a single vertex leaving it convex, like incremental convex hull)
# *** how to deal with very close opposite faces?
# *** if using rounded vertex coordinates, ascertain that the added vertices have integer coordinates too
def split(mesh, splitHeight):
	'''
	Split the mesh into tetra-, pentahedra, one for every face.
	Very close opposite faces (like a concave thin wall) will break visibility info.
	'''
	# we'll return a list of new meshes, together making the original one
	ms = []
	for f in mesh.faces:
		vs = [v.co for v in f.verts]

		# check if the face won't cause trouble
		if len(vs) != 3:
			warning('\t\tSkipping a face not being a triangle, with %d vertices.' % len(vs))
			continue
		if f.area < minFaceArea:
			warning('\t\tSkipping a face with very small area: %.2f.' % f.area)
			continue
		as = filter(lambda a: a < minFaceAngle, BPyMesh.faceAngles(f)) # *** just till the first is found
		if as:
			warning('\t\tSkipping a face with a very small angle: %.2f.' % as[0])
			continue
		es = filter(lambda ei: mesh.edges[ei].length < minEdgeLength, mesh.findEdges(f.edge_keys)) # *** same
		if es:
			warning('\t\tSkipping a face with a very short edge: %.2f.' % mesh.edges[es[0]].length)
			continue

		# make a new tetrahedron, watch vertex order not to flip any normals
		m = Mesh.New()
		m.verts.extend(vs)
		m.verts.extend(f.cent - f.no * splitHeight)
		m.faces.extend(((0, 1, 2), (1, 0, 3), (2, 1, 3), (0, 2, 3)))
		m.faces[0].image = f.image
		m.faces[0].uv = f.uv
		ms.append(m)
	return ms
	'''
	# new faces should have similar area to the original face?
	m.verts.extend(f.cent - f.no * sqrt(f.area) * 1.24 * splitHeight)

	# cuboids, a bit more precise for small faces
	m.verts.extend([v.co - f.no * sqrt(f.area) * splitHeight for v in f.verts])
	m.faces.extend(((0, 1, 2), (1, 0, 3, 4), (2, 1, 4, 5), (0, 2, 5, 3), (5, 4, 3)))
	'''



def Q2Texture(face):
	'''
	Texture mapping info in the Quake 2 format: offset u, offset v, rotation, scale u, scale v;
	all in the one of the base planes (ceiling, west, south etc.); offsets and rotation with respect to the scene center.
	Note that, because textures are actually mapped to one of the base planes and not the face's plane, it's not possible to properly map streched textures,
	or textures to faces when the projections to the base plane of the image edges' in the scene aren't perpendicular.
	Face is assumed to have an image texture, and to be strictly convex.
	*** still requires review, add lack of inverse exception handling
	*** this doesn't work well for textures stretched on quads (completely ignores one vertex)
	'''
	# a plane, the texture is actually mapped to; one of the base planes, the normal of which is the closest to the face normal
	d, (tn, tnnzis) = max(map(lambda p: (dot(face.no, p[0]), p), PLANES))

	# project face points to the base plane (any vertices in any order may be used here) *** quad with three vertices in a line problem
	vs = array([v.co for v in face.verts])[:, tnnzis]
	c = array(face.cent)[tnnzis, ...] # *** c doesn't have to be an array

	# image data, uv coordinates must match vertices
	size = face.image.getSize()
	uvs = array(face.uv)

	# face edges in the scene coordinates and in the image coordinates
	edgesScene = matrix([vs[1] - vs[0], vs[2] - vs[0]])
	edgesImage = matrix([uvs[1] - uvs[0], uvs[2] - uvs[0]])

	# bottom and left edge of the image, or uv unit vectors, in the scene coordinates
	# (uvs may lie on a line -- noninvertible, *** a better approximation of offset and rotation is needed in such cases)
	try:
		units = asarray(edgesImage.I * edgesScene)
	except LinAlgError:
		warning('\t\tA texture line to be mapped onto a face, uvs: %s.' % [tuple(around(v, 2)) for v in face.uv])
		units = asarray(pinv(edgesImage) * edgesScene) # *** does the Moore-Penrose make sense here at all?

	# face center uv coordinates (we'll try to put the same texture point on the face center as blender does)
	try:
		uvc = uvs[0] + ravel(matrix(c - vs[0]) * edgesScene.I * edgesImage)
	except LinAlgError:
		error('\t\tBUG: A line face in texture mapping.')
		return ' 0 0 0 1 1'

	# unscaled scale
	scale = apply_along_axis(norm, 1, units)

	# normalize the unit vectors
	units[0] /= scale[0]
	units[1] /= scale[1]

	# quake understands the left edge upside down, or blender does
	scale[1] *= -1
	# uvs[:, 1] *= -1, not used
	uvc[1] *= -1

	# units and base axes have the same orientation?, if not flip one of the units (projected base axes cross is always -1)
	if cross(units[0], units[1]) > 0:
		units[0] *= -1
		scale[0] *= -1

	# rotation angle (around (0,0,0)), such an angle that the base rotated by it gives the texture units' projections
	# or rather the sum of the base axes gives the sum of the units, any better idea for stretched textures? *** implement the configurable weighting
	rotation = atan2(*ravel(dot(units[0] + units[1], asit)))

	# offsets will be calculated with respect to the rotated base
	r = matrix([[sin(rotation)], [cos(rotation)]])

	# offset (from (0,0,0), in lengths of units' projections)
	# *** shouldn't this be matrix(c - vs[0])? think this over again
	offset = uvc - ravel(matrix(c) * hstack([a0 * r, a1 * r])) / scale

	# in degrees, rounded
	rotation = round(rotation * 180.0 / pi) % 360

	# in images
	scale /= size

	# in images, rounded
	offset = around((offset * size) % size)

	return '%d %d %d %f %f' % (offset[0], offset[1], rotation, scale[0], scale[1])



def writeFace(file, face, flags, coordinatesMultiplier, imagePaths, missingImage):
	'''
	Writes the face description to the file.
	Writes three vertices, texture information (name, u offset, v offset, rotation, u scale, v scale), and flags.
	'''
	# write three points in the face plane (in a counterclock-wise order)
	vs = [face.verts[-1].co, face.verts[1].co, face.verts[0].co]

	# large coordinates give better approximation of plane normal and distance, but an edge shared by two different meshes may come out as two different edges
	if coordinatesMultiplier == 1:
		ps = vs
	else:
		ps = array([vs[1] - vs[0], vs[2] - vs[1], vs[0] - vs[2]], dtype=longdouble) # some increased precision won't hurt here
		ps = apply_along_axis(normalize, 1, ps) * coordinatesMultiplier + vs
	file.write('( %d %d %d ) ( %d %d %d ) ( %d %d %d )' % tuple(around(ravel(ps))))

	# write texture information, it's assumed here that the mesh has an uv layer
	if face.image:
		file.write(' %s %s' % (imagePaths[face.image], Q2Texture(face)))
	else:
		flags = (0, 128, 0) # 16 = ~transparent, 128 = don't draw, 512 = completely skip (breaks visibility)
		file.write(' %s 0 0 0 1 1' % missingImage)

	# write flags, *** all flags handling not just level flags (handling somewhat above or at writeMesh)
	file.write(' %d %d %d\n' % tuple(flags))

	'''
	# write a lot of debugging info to the map file (should still compile)
	file.write('// normal: %s, center: %s, area: %s\n' % (face.no, face.cent, face.area))
	file.write('// vertices: %s\n' % [v.co for v in face.verts])
	file.write('// edges lengths: %f %f %f\n' % (norm(vs[2] - vs[0]), norm(vs[1] - vs[0]), norm(vs[2] - vs[1])))
	file.write('// face angles: %s\n' % str(BPyMesh.faceAngles(face)))
	v1, v2 = normalize(vs[1] - vs[0]), normalize(vs[2] - vs[0])
	pr1, pr2 = normalize(around(ps[1] - ps[0])), normalize(around(ps[2] - ps[0]))
	file.write('// original plane vectors: %s %s, quality: %f %f %f\n' % (v1, v2, dot(v1, face.no), dot(v2, face.no), dot(vs[0], face.no)))
	file.write('// written plane vectors:  %s %s, quality: %f %f %f\n// ---------------------\n' % (pr1, pr2, dot(pr1, face.no), dot(pr2, face.no), dot(ps[0], face.no)))
	'''


def writeMesh(file, mesh, scale, quadToleration, splitMethod, splitHeight, minFaceArea, minFaceAngle, minEdgeLength, coordinatesMultiplier, imagePaths, missingImage, stats):
	'''
	Write a Blender mesh to the map file. It will be split if needed (if it's not strictly concave).
	Writes game logic properties as entity properties, then writes faces of one or more brushes as needed.
	Level visibility is set to above object's geometric center (calculated, not from Blander). You can override this with the 'levels' property.
	'''
	# make a copy of the mesh data
	dm.getFromObject(mesh)
	do.link(dm)

	# *** some more checks to see if at all exportable
	if not dm.faces:
		warning('\t\tIgnoring a mesh without faces.')
		return

	if not dm.faceUV:
		warning('\t\tNo uv layer. Only uv-mapped textures are supported.')
		dm.addUVLayer('Dummy uv layer.')

	# apply object transform to all vertices to get their world coordinates
	dm.transform(mesh.matrix * scale, recalc_normals=True)

	# round vertices before performing any calculations (goes with writing them verbatim later),
	# this can turn some faces completely, however sometimes is better at avoiding breaks in split meshes
	if coordinatesMultiplier == 1:
		for v in dm.verts:
			v.co = Mathutils.Vector(around(v.co))

	# triangulate all nonplanar quads
	dm.sel = False
	for f in dm.faces:
		if len(f.verts) == 4:
			vs = [v.co for v in f.verts]
			es = [vs[i] - vs[(i+1) % 4] for i in range(4)]
			if not alltrue([abs(dot(es[i], f.no)) < quadToleration * norm(es[i]) for i in range(4)]):
				f.sel = True
	dm.quadToTriangle()

	# split the mesh if needed and requested
	if needsSplitting(dm, scale):
		if splitMethod: # *** just one (not really working) method at the moment :)
			ms = split(dm)
		else:
			warning('\t\tSplitting Disabled!')
			ms = [dm]
	else:
		ms = [dm]

	# defined game properties
	ps = dict((p.name, p.data) for p in mesh.game_properties)

	# flags property
	if 'flags' in ps:
		flags = map(int, ps['flags'].split(None, 2))
		del ps['flags']
	else:
		flags = [0, 0, 0]

	# level visibility
	if 'levels' in ps:
		lf = sum(map(lambda x: 1 << x if ps['levels'].find(str(x+1)) != -1 else 0, range(8)))
		stats['levels'] = max(stats['levels'], max(map(int, ps['levels'].split())))
		del ps['levels']
	else:
		vs = dm.verts
		l = int(max(0, min(7, floor(sum(v.co.z for v in vs) / len(vs) / scale / 2))))
		lf = sum(map(lambda x: 1 << x, range(l, 8)))
		stats['levels'] = max(stats['levels'], l+1)
	flags[0] ^= 256 * lf

	# special properties
	if 'actorclip' in ps:
		flags = [65536, 0, 0] # ignore the rest of the flags intentionally
		del ps['actorclip']
	if 'trans33' in ps:
		flags[1] ^= 16
		del ps['trans33']
	if 'trans66' in ps:
		flags[1] ^= 32
		del ps['trans66']

	# write all the brushes needed to render this mesh
	for i in range(len(ms)):
		file.write('// Brush %d (%s %d)\n{\n' % (stats['brushes'], mesh.name, i))
		for p in sorted(ps.iteritems()):
			file.write('"%s" "%s"\n' % p)
		for f in ms[i].faces:
		    writeFace(file, f, flags, coordinatesMultiplier, imagePaths, missingImage)
		file.write('}\n')
		stats['faces'] += len(ms[i].faces)
		stats['brushes'] += 1
	stats['meshes'] += 1


def writeLight(file, light, scale, energyMultiplier, stats):
	'''
	Write a light to the map file.
	Writes game logic properties as entity properties adding light, color and origin if not already defined.
	Light intensity is blenders energy multiplied by the multiplier.
	'''
	# add intensity, color and origin
	ps = dict()
	ps['classname'] = 'light'
	ps['light'] = '%f' % (light.data.energy * energyMultiplier)
	ps['_color'] = '%f %f %f' % tuple(light.data.col)
	ps['origin'] = '%d %d %d' % tuple(around(array(light.loc) * scale))
	ps.update(dict((p.name, p.data) for p in light.game_properties))

	# write the entity
	file.write('// %s\n{\n' % light.name)
	for p in sorted(ps.iteritems()):
		file.write('"%s" "%s"\n' % p)
	file.write('}\n')
	stats['lights'] += 1


def writeModel(file, model, scale, modelsFolder, stats):
	'''
	Writes model entity to the file. Path ("model" property) is required and should begin with models/.
	Adds angles, origin and spawnflags to the defined properties (if not already set).
	You can use "levels" property as with meshes, overriding spawnflags.
	'''
	# check and normalize the path, can be absolute or relative to modelsFolder
	try:
		p = model.getProperty('model')
	except RuntimeError:
		p = None
	if not p:
		warning('\t\tModel does not have path set ("model" property).')
		return
	p = p.data
	if isabs(p):
		p = realpath(p)
		if not p.startswith(modelsFolder):
			warning('\t\tModel is outside of models base: %s.' % p)
			return
	else:
		p = realpath(join(modelsFolder, p))
	if not exists(p):
		warning('\t\tModel doesn\'t exist. Bad path set (%s).' % p)
		return
	p = p[len(modelsFolder)+1:]

	# level -- spawnflags
	l = int(max(0, min(7, floor(model.LocZ / 2))))

	# define origin and angles (pitch, yaw, roll), update with user-defined game properties
	ps = dict()
	ps['origin'] = '%d %d %d' % tuple(around(array(model.loc) * scale))
	ps['angles'] = '%d %d %d' % tuple(around([a * 180.0 / pi for a in (model.RotX, model.RotZ - pi / 2, -model.RotY)]))
	ps['spawnflags'] = sum(map(lambda x: 1 << x, range(l, 8)))
	ps.update(dict((p.name, p.data) for p in model.game_properties))

	# update with the normalized path
	ps['model'] = p

	# override spawnflags with levels if present
	if 'levels' in ps:
		ps['spawnflags'] = sum(map(lambda x: 1 << x if ps['levels'].find(str(x+1)) != -1 else 0, range(8)))
		del ps['levels']

	# write it
	file.write('// %s\n{\n' % model.name)
	for p in sorted(ps.iteritems()):
		file.write('"%s" "%s"\n' % p)
	file.write('}\n')
	stats['levels'] = max(stats['levels'], l+1)
	stats['models'] += 1


def writeEntity(file, entity, scale, stats):
	'''
	Write generic entity to the map file.
	Writes game logic properties as entity properties, adds origin and angle.
	The classname must be already set as game property is required.
	'''
	# add angle and origin
	ps = dict()
	ps['origin'] = '%d %d %d' % tuple(around(array(entity.loc) * scale))
	ps['angle'] = '%d' % round(entity.RotZ * 180.0 / pi)
	ps.update(dict((p.name, p.data) for p in entity.game_properties))

	# write it out, including properties
	file.write('// %s\n{\n' % entity.name)
	for p in sorted(ps.iteritems()):
		file.write('"%s" "%s"\n' % p)
	file.write('}\n')

	# statistics have separate entries for different spawns
	if ps['classname'] == 'info_alien_start':
		stats['aliens'] += 1
	elif ps['classname'] == 'info_human_start':
		stats['humans'] += 1
	elif ps['classname'] == 'info_player_start':
		if not ps['team']:
			warning('Player spawn with no team set.')
		stats['teams'] = max(stats['teams'], ps['team'])
		stats['players'] += 1
	else:
		stats['entities'] += 1



def export(fileName):
	'''
	Exports meshes, lights and all that has classname game property as entities.
	'''
	# logging and printing configuration
	basicConfig(level=15, format='%(message)s', stream=stderr)
	set_printoptions(precision=2, suppress=True)


	# general options
	pathField = Draw.Create(normpath('../ufoai'))
	scaleField = Draw.Create(32.0)
	quadTolerationField = Draw.Create(0.01)
	coordinatesMultiplierField = Draw.Create(65536)

	# light options
	energyMultiplierField = Draw.Create(500)
	sunlightButton = Draw.Create(True)
	sunlightIntensityField = Draw.Create(160)		# day
	sunlightColorField = Draw.Create('1.0 0.8 0.8')	# white
	sunlightAnglesField = Draw.Create('30 210')		# ~noon
	ambientButton = Draw.Create(True)
	ambientColorField = Draw.Create('0.4 0.4 0.4')	# just as example

	# splitting options
	splitButton = Draw.Create(False)
	splitHeightField = Draw.Create(3.5)
	minFaceAreaField = Draw.Create(0.5) 	# 3.5 should be enough in most cases
	minFaceAngleField = Draw.Create(5) 		# 3.5 should be enough in most cases
	minEdgeLengthField = Draw.Create(1) 	# 0.8 usually is enough
	sceneCopyButton = Draw.Create(False)
	leaveDummiesButton = Draw.Create(False)
	faceWarningsButton = Draw.Create(False)

	# ***
	def globalLightButtonCallback(event, value):
		print event, value;

	# that's a popup
	fs = (
	('UFO Path:'),
	('', pathField, 0, 256, 'Path to the top ufo folder.'),
	(' '),
	('General:'),
	('Scale:', scaleField, 0.1, 100.0, 'Scale everything by this value.'),
	('Quad toleration:', quadTolerationField, 0.0, 1.0, 'Consider quad nonplanar if absolute cosine between an edge and normal is higher.'),
	('Coordinates Multiplier:', coordinatesMultiplierField, 1, 131072, 'Enlarge plane coordinates to mitigate rounding errors. Set 1 to disable.'),
	(' '),
	('Light:'),
	('Energy multiplier:', energyMultiplierField, 0, 10000, 'Light intensity for lamps, energy multiplier.'),
	('Sunlight', sunlightButton, 'Enable sunlight.'),
	('Sunlight intensity:', sunlightIntensityField, 0, 10000, 'Global, parallel light intensity.'),
	('Sunlight color:', sunlightColorField, 0, 100, 'Color of the parallel light. Three floats.'),
	('Sunlight angles:', sunlightAnglesField, 0, 100, 'Angles of the parallel light. In degrees.'),
	('Ambient', ambientButton, 'Enable ambient lighting.'),
	('Ambient color:', ambientColorField, 0, 100, 'Ambient lighting color. Three floats.'),
	('Concave splitting:'),
	('Enable', splitButton, 'Make a tetra-, pentahedron for every face, doesn\'t work too well.'),
	('Splitting Height:', splitHeightField, 0.01, 100.0, 'Pyramid height for split faces.'),
	('Min. Face Area:', minFaceAreaField, 0, 100.0, 'Skip faces with area (after scaling) lower than this value.'),
	('Min. Face Angle:', minFaceAngleField, 0, 60.0, 'Skip faces with angles lower than this value (degrees).'),
	('Min. Edge Length:', minEdgeLengthField, 0, 100.0, 'Skip faces with edges shorter than that (after scaling).'),
	(' '),
	(' '),
	('Debugging:'),
	('Copy Scene', sceneCopyButton, 'Operate on a deep scene copy.'),
	('Leave Dummy', leaveDummiesButton, 'Don\'t unlink the dummy object.'),
	('Face Warnings', faceWarningsButton, 'Warn about faces ignored during splitting.'),
	)

	if not Draw.PupBlock('UFO:AI map export', fs):
		return

	Window.WaitCursor(1)

	Window.EditMode(0) # *** reenter if enabled

	st = clock()

	info('Exporting: %s.' % fileName)

	# written things counters
	stats = { 'meshes' : 0, 'lights' : 0, 'models' : 0, 'entities' : 0, 'faces' : 0, 'brushes' : 0, 'aliens' : 0, 'humans' : 0, 'players' : 0, 'teams' : 0, 'levels' : 0 }

	# options
	#	opts = { 'scale' : scaleField.val, 'light' : lightField.val, ...

	# if requested, operate on a deep scene copy
	if sceneCopyButton.val:
		sceneToExport = Scene.GetCurrent()
		scene = sceneToExport.copy(2)
	else:
		scene = Scene.GetCurrent()

	# only textures inside base/textures folder will get exported, only models inside base, all paths will be relative to this base folders
	#*** there should be three 'missing' textures: not texture at all, outside base, split
	texturesFolder = realpath(join(pathField.val, normpath('base/textures')))
	if not exists(texturesFolder):
		error('The textures base folder (%s) does not exist. Incomplete ufo or a bad path.' % texturesFolder)
		return
	modelsFolder = realpath(join(pathField.val, normpath('base')))
	if not exists(modelsFolder):
		error('The models base folder (%s) does not exist. Incomplete ufo or a bad path.' % modelsFolder)
		return
	missingImage = normpath('tex_common/nodraw.tga')
	if not exists(realpath(join(texturesFolder, missingImage))):
		error('The missing texture (%s) is missing.' % missingImage)
		return
	missingImage = splitext(missingImage)[0]

	# preprocess image paths, find paths relative to base/textures, *** how to get images just from the current scene?
	info('Processing images.')
	imagePaths = {}
	for i in Image.Get():
		p = normpath(sys.expandpath(i.getFilename())).split(normpath('base/textures'), 1)
		if len(p) != 2:
			warning('\tAn image outside of the base/textures folder: %s.' % i.getFilename())
			p = missingImage
		else:
			p = p[1]
			if p.startswith(sep):
				p = p[len(sep):]
			p = realpath(join(texturesFolder, p))
			if not exists(p):
				warning('\tAn image doesn\'t exist: %s.' % p)
				p = missingImage
			else:
				p = splitext(p[len(texturesFolder)+1:])[0]
		imagePaths[i] = p

	# warnings about skipped faces are to verbose sometimes
	if not faceWarningsButton.val:
		class FaceWarningsFilter(Filter):
			def filter(self, record):
				return not record.getMessage().startswith('\t\tSkipping a face')
		getLogger().addFilter(FaceWarningsFilter())

	# classify and scale (look for the classname first, then decide by object type)
	info('Classifying.')
	meshes = []
	lights = []
	models = []
	entities = []
	for o in scene.objects:
		if o.name.startswith('Dummy Exporter'):
			warning('\tA probable exception leftover, ignoring: %s.' % o.name)
			continue
		try:
			cn = o.getProperty('classname')
		except RuntimeError: # hail Blender ...
			cn = None
		if cn:
			if cn.data == 'misc_model':
				models.append(o)
			else:
				entities.append(o)
		else:
			if o.type == 'Mesh':
				meshes.append(o)
			elif o.type == 'Lamp':
				lights.append(o)
			else:
				warning('\tIgnoring an object: %s.' % o.name)

	# write to memory, flush later
	world = StringIO()
	rest = StringIO()

	# export meshes
	if meshes:
		info('Exporting meshes:')
		scene.objects.link(do) # dummy object needed for triangulating
		for m in meshes:
			info('\t%s' % m.name)
			writeMesh(world, m, scaleField.val, quadTolerationField.val, splitButton.val, splitHeightField.val, minFaceAreaField.val, minFaceAngleField.val, minEdgeLengthField.val, coordinatesMultiplierField.val, imagePaths, missingImage, stats)
		if not leaveDummiesButton.val:
			scene.objects.unlink(do)

	# export lights
	if lights:
		info('Exporting lights:')
		for l in lights:
			info('\t%s' % l.name)
			writeLight(rest, l, scaleField.val, energyMultiplierField.val, stats)

	# export models
	if models:
		info('Exporting models:')
		for m in models:
			info('\t%s' % m.name)
			writeModel(rest, m, scaleField.val, modelsFolder, stats)

	# export other entities (spawns, models)
	if entities:
		info('Exporting entities:')
		for e in entities:
		    info('\t%s' % e.name)
		    writeEntity(rest, e, scaleField.val, stats)

	# get rid of the scene copy
	if sceneCopyButton.val:
		sceneToExport.makeCurrent()
		Scene.Unlink(scene)

	# check some general things
	if stats['aliens'] == 0:
		warning('Map does not have any alien spawns (classname info_alien_start).')

	if stats['humans'] == 0:
		warning('Map does not have any human spawns (classname info_human_start).')

	if stats['teams'] == 0:
		warning('Map does not have any multiplier spawns (classname info_player_start).')
	# *** check some more, any lights?

	# write to disk
	file = open(fileName, 'wb')
	file.write('// Exported on: %s.\n' % datetime.now())
	file.write('{\n')
	file.write('"classname" "worldspawn"\n')
	if sunlightButton.val:
		file.write('"light" "%d"\n' % sunlightIntensityField.val) # *** check light values
		file.write('"_color" "%s"\n' % sunlightColorField.val)
		file.write('"angles" "%s"\n' % sunlightAnglesField.val)
	if ambientButton.val:
		file.write('"ambient" "%s"\n' % ambientColorField.val)
	file.write('"maxlevel" "%d"\n' % (stats['levels'] + 1)) # need to add one in case of actor getting to the map top
	file.write('"maxteams" "%d"\n' % stats['teams'])
	file.write(world.getvalue())
	file.write('}\n')
	file.write(rest.getvalue())
	file.close()

	# bah
	s = lambda n: [n, '' if n == 1 else 's']
	es = lambda n: [n, '' if n == 1 else 'es']
	ies = lambda n: [n, 'y' if n == 1 else 'ies']
	info(str(stats))
	info('%d mesh%s, %d light%s, %d model%s, %d other entit%s, done in %.2f second%s.' % tuple(es(stats['meshes']) + s(stats['lights']) + s(stats['models']) + ies(stats['entities']) + s(clock() - st)))

	Window.WaitCursor(0)



def main():
	# *** some default reasonable for everyone
	Window.FileSelector(export, 'UFO:AI map export', realpath(normpath('../ufoai/base/maps/blenderd.map')))
	'''
	# profiling, open stats with pprofui for example
	from profile import run
	from os.path import expanduser
	run("from os.path import *; from ufoai_export import export; export(realpath(normpath('../ufoai/base/maps/blender.map')))", expanduser('~/stats'))
	'''

if __name__ == '__main__': main()



# Print all marker names

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C): DarkRain, Werner Höhrer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------

import bpy
import bpy.ops

#=======================
# Collect marker data
#=======================
def generate_mlist (markerData, printInfo):
	if (printInfo):
		print("Frame: Marker name")

	# Get the list of markers (includes start-frame and name)
	frameNames =[]
	for marker in bpy.context.scene.timeline_markers:
		frameNames.append(marker)
	frameNames.sort(key=lambda marker: marker.frame)
#	print(frameNames) #debug

	previousFrame = -1
	for i in range(len(frameNames)):
		frame = frameNames[i].frame
		if (previousFrame < 0):
			previousFrame = frame

		name = frameNames[i].name
		if (printInfo):
			print(str(frame) + ": " + name)
		# Set data (mainly name) of current frame
		markerData[frame] = (name, frame);
		# Set end-frame of previous frame
		markerData[previousFrame] = (markerData[previousFrame][0], frame-1)
		# Remember the previous frame
		previousFrame = frame

	if (printInfo):
		print(str(frame) + ": Last frame.")
		print(str(len(markerData)) + " markers found.")


#=======================
# Output ufoai data
#=======================
def write_mlist_ufoai (markerData):
	bpy.data.texts.new(name="framelist")
	txt = bpy.data.texts["framelist"]

	#file = open(filename, 'wb')
	try:
		ufoaiOffset = bpy.context.scene.frame_start	# Difference between Blender/UFO:AI frames (ufoai = blender - ufoaiOffset)

#		print(markerData)#debug
		dummySpeed = 20
		txt.write("// Frame list exported from blender - please check if everything is ok before using!\n")
		for frame in sorted(markerData.keys()):
			if (markerData[frame][0] == "ufoai_end"):
				break
			txt.write(markerData[frame][0] + "\t\t\t\t" + str(frame-ufoaiOffset) + "\t\t" + str(markerData[frame][1]-ufoaiOffset) + "\t\t" + str(dummySpeed) + "\n")
		txt.write("// End of the list.\n")
	finally:
		print("========================================")
		for line in txt.lines:
			print(line.body)
		print("========================================")
		print("Created new text with the frame list. See text-editor for saving.")
		print("A badly formatted version is printed above.")
		# You can normally use [Ctrl][Shift][Alt][C] to copy the text into the clipboards for use in other applications. Might not always work though.
		#file.close()

#=======================
# MAIN
#=======================
ufoaiOutput = True
mlist = {}
generate_mlist(mlist, not ufoaiOutput)

if ufoaiOutput:
	write_mlist_ufoai(mlist)



#! /usr/bin/env python

# Compile src/po/*.po to base/i18n using command-line supplied "msgmerge" path
# This is primarily for windows users. "msgmerge" is usually in the same place
# as poedit
#
# by techtonik // rainforce.org

import os, glob, sys, stat, platform

#read the given path
if len(sys.argv) < 2:
    msgfmt = "msgfmt"
    if platform.system() == "Windows": msgfmt = msgfmt + ".exe"
else:
    msgfmt = sys.argv[1]

print "====================================================="
print "po compiler for UFO:AI (http://sf.net/projects/ufoai)"
print "Given path to 'msgfmt' utility as parameter it will  "
print "compile all .po files and in src/po and place them   "
print "into base/i18n/"
print "====================================================="

sys.stdout.write("...checking for msgfmt: "); sys.stdout.flush();
if not os.access(msgfmt, os.X_OK):
    print "is not executable. exiting..."
    print
    print "usage: compile_po.py E:\\path\\to\\poEdit\\bin\\msgfmt.exe"
    sys.exit(1)
print "%s - ok" % msgfmt

podir = "."
podir = os.path.normpath(podir)
if not os.path.isdir(podir):
    print "%s does not exists" % podir
    sys.exit(1)
i18ndir = "../../base/i18n"
i18ndir = os.path.normpath(i18ndir)
if not os.path.isdir(i18ndir):
    print "%s does not exists" % i18ndir
    sys.exit(1)

po = glob.glob(podir + "/*.po")
for f in po:
    name, ext = os.path.splitext(os.path.split(f)[1])
    outpath = os.path.normpath(i18ndir + "/" + name + "/LC_MESSAGES")
    if not os.path.isdir(outpath):
        os.makedirs(outpath)
    outfile = os.path.normpath(outpath + "/" + "ufoai.mo")
    sys.stdout.write("compiling %s to %s -" % (f, outfile)); sys.stdout.flush();

    z = os.system("%s -o %s %s" % (msgfmt, outfile, f))
    if z != 0:
        sys.exit(" msgfmt error. Terminated.")

    print "ok"

print "Done."



#!/usr/bin/env python
# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-

# Copyright (C) 2011 - UFO: Alien Invasion
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import dircache
import os.path
import Image
import sys

# TODO check all icons from a set must have same size
# TODO merge similar sprites

# root of the game
UFOAI_ROOT = os.path.realpath(sys.path[0] + '/../..')

SUFFIXES = set(["_clicked", "_hovered", "_disabled"])

#####################################################

class Sprite:
	def __init__(self, url):
		self.image = Image.open(url)
		self.filename = url
		self.clickedIcon = None
		self.hoveredIcon = None
		self.disabledIcon = None
		self.pack = None
		self.pos = None

		base = os.path.splitext(url)[0]
		extension = os.path.splitext(url)[1]

		url = base + "_clicked" + extension
		if os.path.exists(url):
			self.clickedIcon = Sprite(url)
		url = base + "_hovered" + extension
		if os.path.exists(url):
			self.hoveredIcon = Sprite(url)
		url = base + "_disabled" + extension
		if os.path.exists(url):
			self.disabledIcon = Sprite(url)

		self.name = base.split('/pics/')[-1]

	def getIconSet(self):
		icons = set([])
		icons.add(self)
		if self.clickedIcon:
			icons.add(self.clickedIcon)
		if self.hoveredIcon:
			icons.add(self.hoveredIcon)
		if self.disabledIcon:
			icons.add(self.disabledIcon)
		return icons

	def size(self):
		return self.image.size
	def width(self):
		return self.image.size[0]
	def height(self):
		return self.image.size[1]

class Pack:
	def __init__(self, name, size, margin = 1):
		transparent = (0, 0, 0, 0)
		self.image = Image.new("RGBA", [size, size], transparent)
		self.width = size
		self.height = size
		self.availableHeight = size - margin - margin
		self.availableWidth = size - margin - margin
		self.margin = margin
		self.rowHeight = 0
		self.name = name
        self.icons = set([])

	# return available position for a new icon
	def getAvailableIconPosition(self, icon):
		pos = None
		newRow = False

		# append to a row
		if self.availableWidth != None:
			if icon.width() <= self.availableWidth:
				if icon.height() > self.availableHeight:
					return None
				pos = self.width - self.availableWidth - self.margin, self.height - self.availableHeight - self.margin
			else:
				newRow = True

		# start a new row
		if newRow:
			if icon.height() > self.availableHeight - self.rowHeight:
				return None
			if icon.width() > self.width - self.margin - self.margin:
				return None
			pos = self.margin, self.height - self.availableHeight + self.rowHeight - self.margin

		return pos

	def _updatePackCacheToAddIcon(self, pos, icon):
		# do we start a new row
		if pos[0] == self.margin:
			self.availableWidth = self.width - self.margin - self.margin
			if self.rowHeight != 0:
				self.availableHeight -= self.rowHeight + self.margin
			self.rowHeight = 0
		if icon.height() > self.rowHeight:
			self.rowHeight = icon.height()
		self.availableWidth -= icon.width() + self.margin

	def _addIcon(self, icon, test):
		availableWidth = self.availableWidth
		availableHeight = self.availableHeight
		rowHeight = self.rowHeight

		icons = icon.getIconSet()

		added = False
		for i in icons:
			pos = self.getAvailableIconPosition(i)
			added = pos != None
			if added:
				self._updatePackCacheToAddIcon(pos, i)
                self.icons.add(i)
				if not test:
					self.image.paste(i.image, pos)
					i.pos = pos
					i.pack = self
			else:
				break

		if test:
			self.availableWidth = availableWidth
			self.availableHeight = availableHeight
			self.rowHeight = rowHeight

		return added

	# test to add an icon, return True if the icon is added
	# the function do not change the internal state of the object
	def testAddIcon(self, icon):
		return self._addIcon(icon, True)

	# add an icon, return True if the icon is added
	def addIcon(self, icon):
		test = self.testAddIcon(icon)
		if not test:
			return
		return self._addIcon(icon, False)

def loadIcons(srcDir, options):
	global verbose
    print 'Load sprites from \"%s\"' % srcDir
	result = []
	list = dircache.listdir(srcDir)
	iconCount = 0
	iconsetCount = 0
	for filename in list:
		# filter extension
		extension = os.path.splitext(filename)[1]
		extension = extension.lower()
		if not (extension in '.png|.jpg|.jpeg|.tga|.gif'):
			print >> sys.stderr, "Warning: Skip \"%s\"" % filename
			continue

		# avoid pack files
		if "pack" in filename:
			print >> sys.stderr, "Warning: Skip \"%s\"" % filename
			continue

		# filter special suffixes
		name = os.path.splitext(filename)[0]
		special = False
		for suffix in SUFFIXES:
			if name.endswith(suffix):
				special = True
				break
		if special:
			continue

		icon = Sprite(srcDir + '/' + filename)
		if options.verbose:
			print filename, icon.size(), "x" + str(len(icon.getIconSet()))
		iconCount += len(icon.getIconSet())
		iconsetCount += 1
		result.append(icon)
	print "    Icons count: ", iconCount
	print "    Iconsets count: ", iconsetCount
	return result

def iconsByHeight(x, y):
        return x.height() - y.height()

def iconsByName(x, y):
        return cmp(x.name, y.name)

def generateIconPacks(iconList, packSize, options):
	list = iconList[:]
	list.sort(cmp=iconsByHeight)
	packs = []

	while len(list) > 0:
		icon = list.pop()
		added = False
		for p in packs:
			added = p.addIcon(icon)
			if options.verbose:
				print icon.filename + " -> " + p.name, added
			if added:
				break
		if not added:
			p = Pack("autogen" + str(len(packs)), packSize, margin=options.margin)
			packs.append(p)
			added = p.addIcon(icon)
			if not added:
				raise Exception("No way to add this icon to a new pack")
	print "Iconpacks count: ", len(packs)

	return packs

def savePacks(packs, destDir, format):
	if not os.path.exists(destDir):
		os.makedirs(destDir)
	for p in packs:
		name = p.name + "." + format.lower()
        fileName = destDir + "/" + name
        print "Generating %s..." % fileName
		p.image.save(fileName, format)

        fileName = fileName + ".license"
        print "Generating %s..." % fileName
        file = open(fileName, "wt")
        file.write("This file is compouned by many single image from many authors and licenses.\n")
        file.write("The file ./base/ufos/sprites_autogen.ufo identify each zone of this image.\n")
        file.write("and ./LICENSES describe all author/license/source for all this identifiers.\n")
        file.write("All this sprites are available in single on the source repo.\n")
        file.write("Contact us for more information.\n")
        file.close()

def generateIconScript(icons, destDir):
	if not os.path.exists(destDir):
		os.makedirs(destDir)

    fileName = destDir + '/sprites_autogen.ufo'
    print "Generating %s..." % fileName
	script = open(fileName, "wt")
	script.write("// It is an autogenerated file from ./src/resources/sprite_pack.py\n")
	script.write("\n")

	icons = icons[:]
	icons.sort(cmp=iconsByName)
	for i in icons:
		script.write("sprite " + i.name + " {\n")
		script.write("\timage \"banks/" + i.pack.name + "\"\n")
		script.write("\tsize \"" + str(i.width()) + " " + str(i.height()) + "\"\n")
		script.write("\ttexl \"" + str(i.pos[0]) + " " + str(i.pos[1]) + "\"\n")
		if i.hoveredIcon:
			script.write("\thoveredtexl \"" + str(i.hoveredIcon.pos[0]) + " " + str(i.hoveredIcon.pos[1]) + "\"\n")
		if i.disabledIcon:
			script.write("\tdisabledtexl \"" + str(i.disabledIcon.pos[0]) + " " + str(i.disabledIcon.pos[1]) + "\"\n")
		if i.clickedIcon:
			script.write("\tclickedtexl \"" + str(i.clickedIcon.pos[0]) + " " + str(i.clickedIcon.pos[1]) + "\"\n")

		script.write("}\n")
	script.close()

#####################################################

def gen_sprite_pack(base_destination, options, remove_source):
    pack_size = options.pack_size
    sprites = []
    for src in options.source_dir:
        sprites += loadIcons(src, options)
    packs = generateIconPacks(sprites, pack_size, options)
    pack_dest = os.path.join(base_destination, "pics", "banks")
    savePacks(packs, pack_dest, options.pack_format)
    script_dest = os.path.join(base_destination, "ufos")
    generateIconScript(sprites, script_dest)

    if remove_source:
        count = 0
        for pack in packs:
            count += len(pack.icons)

        print "Remove source file: %i sprites" % count
        for pack in packs:
            for sprite in pack.icons:
                if sprite.pack == None:
                    continue
                if options.verbose:
                    print "#   " + sprite.filename
                os.remove(sprite.filename)

def command_clean(options):
    base_dest = os.path.join(UFOAI_ROOT, options.base_dest)
    import glob
    remove_list = glob.glob(os.path.join(base_dest, "pics", "banks", "autogen*"))

    file = os.path.join(base_dest, "ufos", "sprites_autogen.ufo")
    if os.path.exists(file):
        remove_list.append(file)

    if len(remove_list) > 0:
        print "Delete files:"
        for file in remove_list:
            print file
            os.remove(file)
    else:
        print "Nothing to clean"

def command_info(options):
    print "Information:"
    print "Format of the generated textures:  %s" % options.pack_format
    print "Size of the generated textures:    %i" % options.pack_size
    print "Margin between sprites:            %i" % options.margin
    print "Sources of sprites:"
    for source in options.source_dir:
        print "#   %s" % source
    print "Destination base of files          %s" % options.base_dest

def main():
    from optparse import OptionParser

    parser = OptionParser(usage="usage: %prog [OPTIONS] COMMAND [ARGS]", add_help_option=False)
    parser.add_option("", "--size", dest="pack_size",
                      help="Size of the generated textures", type="int", default=512)
    parser.add_option("", "--margin", dest="margin",
                      help="Margin between sprites", type="int", default=1)
    parser.add_option("", "--format", dest="pack_format",
                      help="Format of the generated textures", type="string", default="PNG")
    parser.add_option("", "--base-dest", dest="base_dest",
                      help="Base destination for the generated files", type="string", default="base")
    parser.add_option("-h", "--help", dest="help",
                      help="Display this screen", action="store_true", default=False)
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False,
                      help="Print more information to stdout")

    (options, args) = parser.parse_args()

    # where are single sprites
    options.source_dir = [
        os.path.join(UFOAI_ROOT, "base/pics/icons"),
        os.path.join(UFOAI_ROOT, "base/pics/buttons"),
    ]

    if len(args) < 1 or options.help or args[0] == "help":
        parser.print_help()
        print """
Commands:
   generate             Generate atlas textures of icons and expected scripts
   clean                Delete generated atlas files
   release              Generate altas textures and delete sources
                        files. Attention, is
   info                 Display information about the process
   help                 Display this screen
"""
        exit(0)

    if args[0] == "generate":
        base_dest = os.path.join(UFOAI_ROOT, options.base_dest)
        gen_sprite_pack(base_dest, options, False)
    elif args[0] == "clean":
        command_clean(options)
    elif args[0] == "release":
        base_dest = os.path.join(UFOAI_ROOT, options.base_dest)
        gen_sprite_pack(base_dest, options, True)
    elif args[0] == "info":
        command_info(options)

if __name__ == '__main__':
    main()



# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****

'''
This script exports a Quake 2 tag file as used in UFO:AI (md2 TAGs).

This file format equals the _parts_ of the MD3 format used for animated tags, but nothing more. i.e tagnames+positions

Base code taken from the md2 exporter by Bob Holcomb. Many thanks.
'''

bl_info = {
	"name": "UFO:AI md2 TAGs (.tag)",
	"description": "Export to md2 TAG file format used by UFO:AI (.tag).",
	"author": "DarkRain, based on the work of Werner Hoehrer and Bob Holcomb",
	"version": (1, 0),
	"blender": (2, 63, 0),
	"location": "File > Import/Export > UFO:AI md2 TAGs",
	"warning": '', # used for warning icon and text in addons panel
	"support": 'COMMUNITY',
	"category": "Import-Export"}

import bpy
from bpy.props import *
from bpy_extras.io_utils import ExportHelper, ImportHelper
import mathutils

import struct

MD2_MAX_TAGS = 4096
MD2_MAX_FRAMES = 1024

def asciiz(s):
	for i, c in enumerate(s):
		if ord(c) == 0:
			return s[:i]

class Md2TagName:
	name = ''
	format = "<64s"

	def __init__(self, string):
		self.name = string

	def getSize(self):
		return struct.calcsize(self.format)

	def save(self, file):
		data = struct.pack(self.format, bytes(self.name[0:63], encoding="utf-8"))
		file.write(data)

	def load(self, file):
		buff = file.read(self.getSize())
		data = struct.unpack(self.format, buff)
		self.name = asciiz(data[0].decode("utf-8", "replace"))
		return self

class Md2Tag:
	axis1 = []
	axis2 = []
	axis3 = []
	origin = []

	format = "<12f"	#little-endian (<), 12 floats (12f)	| See http://docs.python.org/lib/module-struct.html for more info.

	def __init__(self):
		origin = mathutils.Vector((0.0, 0.0, 0.0))
		axis1 = mathutils.Vector((0.0, 0.0, 0.0))
		axis2 = mathutils.Vector((0.0, 0.0, 0.0))
		axis3 = mathutils.Vector((0.0, 0.0, 0.0))

	def getSize(self):
		return struct.calcsize(self.format)

	# Function + general data-structure stolen from https://intranet.dcc.ufba.br/svn/indigente/trunk/scripts/md3_top.py
	# Kudos to Bob Holcomb and Damien McGinnes

	def save(self, file):
		# Prepare serialised data for export with transformed axes (matrix transform)
		data = struct.pack(self.format,
							# rotation matrix
							float(self.axis2[1]), -float(self.axis2[0]), -float(self.axis2[2]),
							-float(self.axis1[1]), float(self.axis1[0]), float(self.axis1[2]),
							-float(self.axis3[1]), float(self.axis3[0]), float(self.axis3[2]),
							# position
							-float(self.origin[1]), float(self.origin[0]), float(self.origin[2])
							)

		# Write it
		file.write(data)

	def load(self, file):
		# Read data from file with the defined size.
		buff = file.read(self.getSize())

		# De-serialize the data.
		data = struct.unpack(self.format, buff)

		# Parse data from file into python md2 structure
		self.axis2 = (-data[1], data[0], -data[2])
		self.axis1 = (data[4], -data[3], data[5])
		self.axis3 = (data[7], -data[6], data[8])
		self.origin = (data[10], -data[9], data[11])

		return self

class Md2TagsObj:
	#Header Structure
	ident = 0		#int 0	This is used to identify the file
	version = 0		#int 1	The version number of the file (Must be 1)
	numTags = 0		#int 2	The number of tags associated with the model
	numFrames = 0		#int 3	The number of animation frames

	ofsNames = 0	#int 4	The offset in the file for the name data
	ofsTags = 0 	#int 5	The offset in the file for the tags data
	ofsEnd = 0		#int 6	The end of the file offset
	ofsExtractEnd = 0	#int 7	???
	format = "<8i" 	#little-endian (<), 8 integers (8i)

	#md2 tag data objects
	names = []
	tags = [] # (consists of a number of frames)

	'''
	"tags" example:
	tags = (
			(tag1_frame1, tag1_frame2, tag1_frame3, etc...),	# tagFrames1
			(tag2_frame1, tag2_frame2, tag2_frame3, etc...),	# tagFrames2
			(tag3_frame1, tag3_frame2, tag3_frame3, etc...)		# tagFrames3
			)
	'''

	def __init__ (self):
		self.names = []
		self.tags = []

	def getSize(self):
		return struct.calcsize(self.format)

	def save(self, file):
		data = struct.pack(self.format,
							self.ident,
							self.version,
							self.numTags,
							self.numFrames,
							self.ofsNames,
							self.ofsTags,
							self.ofsEnd,
							self.ofsExtractEnd,
							)
		file.write(data)

		#write the names data
		for name in self.names:
			name.save(file)
		tagCount = 0
		for tagFrames in self.tags:
			tagCount += 1
			frameCount = 0
			print("Saving tag: ", tagCount)

			for tag in tagFrames:
				frameCount += 1
				tag.save(file)
			print("Frames saved: ", frameCount, "(",self.numFrames, ")")

	def load(self, file):
		buff = file.read(self.getSize())
		data = struct.unpack(self.format, buff)

		self.ident = data[0]
		self.version = data[1]

		if (self.ident != 844121162 or self.version != 1):
			raise NameError("Not a valid MD2 TAG file")

		self.numTags = data[2]
		self.numFrames = data[3]
		self.ofsNames = data[4]
		self.ofsTags = data[5]

		# Read names data
		file.seek(self.ofsNames, 0)
		for tagCounter in range(self.numTags):
			tempName = Md2TagName("")
			self.names.append(tempName.load(file))
		print("Reading of ", tagCounter, " names finished.")

		file.seek(self.ofsTags, 0)
		for tagCounter in range(self.numTags):
			frames = []
			for frameCounter in range(self.numFrames):
				tempTag = Md2Tag()
				tempTag.load(file)
				frames.append(tempTag)
			self.tags.append(frames)
			print("Reading of ", frameCounter + 1, " frames finished.")

		print("Reading of ", tagCounter + 1, " framelists finished.")
		return self

def outFrame(tagFrames, matrix, frameCounter):
	tagFrames.append(Md2Tag())

	# Set first coordiantes to the location of the empty.
	tagFrames[frameCounter].origin = matrix.translation.copy()
	tagFrames[frameCounter].axis1 = matrix.col[0].copy()
	tagFrames[frameCounter].axis2 = matrix.col[1].copy()
	tagFrames[frameCounter].axis3 = matrix.col[2].copy()

def inFrame(blenderTag, frame):
	#bpy.context.scene.frame_set(frameCount)
	transform = mathutils.Matrix(((frame.axis1[0], frame.axis2[0], frame.axis3[0], frame.origin[0]),
					  (frame.axis1[1], frame.axis2[1], frame.axis3[1], frame.origin[1]),
					  (frame.axis1[2], frame.axis2[2], frame.axis3[2], frame.origin[2]),
					  (0.0, 0.0, 0.0, 1.0)))
	blenderTag.matrix_world = transform

def fillMd2Tags(md2Tags, object, options):
	# Set header information.
	md2Tags.ident = 844121162
	md2Tags.version = 1
	md2Tags.numTags += 1
	frameCounter = 0

	# Add a name node to the tagnames data structure.
	md2Tags.names.append(Md2TagName(object.name))

	# Add a (empty) list of tags-positions (for each frame).
	tagFrames = []

	if options.bExportAnimation:
		# Store currently set frame
		previousCurFrame = bpy.context.scene.frame_current

		# Fill in each tag with its positions per frame
		print("Blender startframe: ", bpy.context.scene.frame_start)
		print("Blender endframe: ", bpy.context.scene.frame_end)
		for currentFrame in range(bpy.context.scene.frame_start , bpy.context.scene.frame_end + 1):
			#set blender to the correct frame (so the objects have their new positions)
			bpy.context.scene.frame_set(currentFrame)
			#add a frame
			outFrame(tagFrames, object.matrix_world, frameCounter)
			frameCounter += 1
		# Restore curframe from before the calculation.
		bpy.context.scene.frame_set(previousCurFrame)
	else:
		outFrame(tagFrames, object.matrix_world, frameCounter)

	md2Tags.tags.append(tagFrames)

class Export_TAG(bpy.types.Operator, ExportHelper):
	'''Export to md2 TAG format used by UFO:AI (.tag)'''
	bl_idname = "export_md2.tag"
	bl_label = "Export to md2 TAG format used by UFO:AI (.tag)"
	filename_ext = ".tag"

	bExportAnimation = BoolProperty(name="Export animation",
					description="default: False",
					default=False)

	def __init__(self):
		try:
			self.objects = bpy.context.selected_objects
		except:
			self.objects = None

	######################################################
	# Save md2 TAGs Format
	######################################################
	def execute(self, context):
		filePath = self.filepath
		filePath = bpy.path.ensure_ext(filePath, self.filename_ext)
		md2Tags = Md2TagsObj()

		md2Tags.numFrames = 1
		if self.bExportAnimation:
			md2Tags.numFrames = 1 + bpy.context.scene.frame_end - bpy.context.scene.frame_start

		if len(bpy.context.selected_objects) == 0:
			raise NameError("Please, select one object!")
		elif len(bpy.context.selected_objects) > MD2_MAX_TAGS:
			raise NameError("Too many objects (%i) selected, at most %i tags supported" % (len(bpy.context.selected_objects), MD2_MAX_TAGS))
		if md2Tags.numFrames > MD2_MAX_FRAMES:
			raise NameError("Too many frames (%i), at most %i are supported" % (md2Tags.numFrames, MD2_MAX_TAGS))

		# go into object mode before we start the actual export procedure
		bpy.ops.object.mode_set( mode='OBJECT', toggle=False )

		for object in self.objects:
			#check if it's an "Empty" mesh object
			if object.type != 'EMPTY':
				print("Ignoring non-'Empty' object: " + object.type)
			else:
				print("Found Empty: " + object.name)
				fillMd2Tags(md2Tags, object, self)

		# Set offset of names
		tempHeader = Md2TagsObj();
		md2Tags.ofsNames = tempHeader.getSize();

		# Set offset of tags
		tempName = Md2TagName('');
		md2Tags.ofsTags = md2Tags.ofsNames + tempName.getSize() * md2Tags.numTags;

		# Set EOF offest value.
		tempTag = Md2Tag();
		md2Tags.ofsEnd = md2Tags.ofsTags + (tempTag.getSize() * md2Tags.numFrames * md2Tags.numTags)
		md2Tags.ofsExtractEnd = md2Tags.ofsTags + (64 * md2Tags.numFrames * md2Tags.numTags)

		# Actually write it to disk.
		fileExport = open(filePath, "wb")
		try:
			md2Tags.save(fileExport)
		finally:
			fileExport.close()

		# Cleanup
		md2Tags = 0

		return {'FINISHED'}

	def invoke(self, context, event):
		wm = context.window_manager
		wm.fileselect_add(self)
		return {'RUNNING_MODAL'}

class Import_TAG(bpy.types.Operator, ImportHelper):
	'''Import from md2 TAG format used by UFO:AI (.tag)'''
	bl_idname = "import_md2.tag"
	bl_label = "Import from md2 TAG format used by UFO:AI (.tag)"

	bImportAnimation = BoolProperty(name="Import animation",
					description="Import all frames (default: False)",
					default=False)

	def __init__(self):
		if len(bpy.context.selected_objects) > 0:
			# go into object mode before we start the actual import procedure
			bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

	######################################################
	# Load md2 TAGs Format
	######################################################
	def execute(self, context):
		md2Tags = Md2TagsObj()

		# deselect all objects
		bpy.ops.object.select_all(action='DESELECT')

		fileImport = open(self.filepath, "rb")
		try:
			md2Tags.load(fileImport)
		finally:
			fileImport.close()

		scene = bpy.context.scene

		for tag in range(md2Tags.numTags):
			# Get name & frame-data of current tag.
			name = md2Tags.names[tag]

			# Create object.
			blenderTag = bpy.data.objects.new(name=name.name, object_data=None)

			# Activate name-visibility for this object.
			blenderTag.show_name = True

			# Link Object to current Scene
			scene.objects.link(blenderTag)
			blenderTag.select = True
			scene.objects.active = blenderTag

			if self.bImportAnimation:
				for i, frame in enumerate(md2Tags.tags[tag]):
					inFrame(blenderTag, frame)
					# Insert keyframe.
					blenderTag.keyframe_insert(data_path='location', frame=i + 1)
					blenderTag.keyframe_insert(data_path='rotation_euler', frame=i + 1)
					blenderTag.keyframe_insert(data_path='scale', frame=i + 1)
			else:
				inFrame(blenderTag, md2Tags.tags[tag][0])

		scene.frame_set(scene.frame_current)
		scene.update()
		return {'FINISHED'}

	def invoke(self, context, event):
		wm = context.window_manager
		wm.fileselect_add(self)
		return {'RUNNING_MODAL'}

def menu_func_export(self, context):
	self.layout.operator(Export_TAG.bl_idname, text="UFO:AI md2 TAGs (.tag)")

def menu_func_import(self, context):
	self.layout.operator(Import_TAG.bl_idname, text="UFO:AI md2 TAGs (.tag)")

def register():
	bpy.utils.register_module(__name__)
	bpy.types.INFO_MT_file_export.append(menu_func_export)
	bpy.types.INFO_MT_file_import.append(menu_func_import)

def unregister():
	bpy.utils.unregister_module(__name__)
	bpy.types.INFO_MT_file_export.remove(menu_func_export)
	bpy.types.INFO_MT_file_import.remove(menu_func_import)

if __name__ == "__main__":
	register()



#!/usr/bin/env python
# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-

import os
import urllib2
import socket
import platform
import sys
import time
import re

VERSION = "1"
RETRIES = 10
UFOAI_ROOT = os.path.realpath(sys.path[0] + '/../..')
POTTLE_BASE = "/pootle"
POTTLE_URL = "http://ufoai.org" + POTTLE_BASE


displayDownloadStatus = True

ignore_languages = set(['en', 'en_BASE'])

# code name from pottle to ufoai file
# TODO: remove it when it is possible
language_mapping = {
    "bg":   "bg_BG",
    "nb":   "no",
    "nl":   "nl_NL",
}

#read the given path

def download(uri):
    global displayDownloadStatus
    request = urllib2.Request(uri)

    p = ' '.join([platform.platform()] + list(platform.dist()))
    request.add_header('User-Agent', 'ufoai_get_po/%s (%s)' % (VERSION, p))

    trynum = 0
    while trynum < RETRIES:
        try:
            f = urllib2.build_opener().open(request)

            re = out = ''
            t = 1
            data = f.read(1024)
            if not displayDownloadStatus:
                print('Downloading ' + uri)
            while data:
                re+= data
                if displayDownloadStatus and sys.stdout.isatty:
                    out = '\r%s %9i KiB' % (uri, len(re) / 1024)
                    sys.stdout.write(out)
                    sys.stdout.flush()
                t = time.time()
                data = f.read(1024)
                t = time.time() - t
            f.close()
            if displayDownloadStatus:
                sys.stdout.write('\r%s\r' % (' '*len(out)))
            break
        except socket.timeout:
            print '...timeout fetching URL'
            trynum += 1
            if trynum >= RETRIES:
                raise
    return re

def find_available_languages():
    data = download(POTTLE_URL + "/projects/ufoai/")
    languages = re.findall("%s/([a-zA-Z_]*)/ufoai/" % POTTLE_BASE, data)
    languages = set(languages)
    languages.remove('projects')
    return languages

def get_language(lang):
    global language_mapping
    url = "%s/%s/ufoai/ufoai.po/download/" % (POTTLE_URL, lang)

    file_lang = lang
    if lang in language_mapping:
        file_lang = language_mapping[lang]
    file_name = "%s/src/po/ufoai-%s.po" % (UFOAI_ROOT, file_lang)

    print 'Download "%s": %s' % (lang, url)
    data = download(url)

    print 'Write "%s": %s' % (lang, file_name)
    file = open(file_name, "wt")
    file.write(data)
    file.close()

def get_all_languages():
    global ignore_languages
    languages = find_available_languages()
    print "Available languages: " + ', '.join(languages)
    for lang in languages:
        if lang in ignore_languages:
            print 'Skip "%s"' % (lang)
            continue
        get_language(lang)

get_all_languages()



#!BPY

"""
Name: 'UFORadiant Prism/Tetra-based (.map)'
Blender: 249
Group: 'Export'
Tooltip: 'Export terrain and triangulated cubed into UFO Radiant map format'
"""

__author__ = 'Martin Meister'
__version__ = '0.1a'
__email__ = "ma_meister@web.de"
__bpydoc__ = """\
This script Exports a blender map made from cubes with triangulated side into UFO Radiant map format as well as prisms for height maps.

 Supports only cubes with triangulated sides and lights
"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C): Campbell Barton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------

# ------------------------------------------------
# 2011-02-01
# THis is modelled after
# "Campbell Barton's Quake .map exporter"
# thanks!
# ---------------------------------------------


from Blender import *
import BPyMesh

PREF_SCALE= Draw.Create(1)
PREF_FACE_THICK= Draw.Create(1024)
PREF_GRID_SNAP= Draw.Create(1)
# Quake 1/2?
# PREF_DEF_TEX_OPTS= Draw.Create(' 0 0 0 1 1\n') # not user settable yet
# Quake 3+?

#PREF_DEF_TEX_OPTS= Draw.Create(' 0 0 0 1 1 0 0 0\n') # not user settable yet
PREF_DEF_TEX_OPTS = Draw.Create(' 0 0 0 1 1 0 128  0\n')
PREF_DEF_TERR_OPTS= Draw.Create(' 0 0 0 1 1 0 1024 0\n')

#PREF_NULL_TEX= Draw.Create('NULL') # not user settable yet
#PREF_INVIS_TEX= Draw.Create('common/caulk')

PREF_TERR_TEX= Draw.Create('tex_nature/dirt')
PREF_INVIS_TEX= Draw.Create('tex_common/nodraw')


# prism planar projection center
PREF_PROJ_X = Draw.Create(0.0)
PREF_PROJ_Y = Draw.Create(0.0)
PREF_PROJ_Z = Draw.Create(-1.0)

def write_cube2brush(file, faces):
	'''
	Takes 6 faces and writes a brush,
	these faces can be from 1 mesh, 1 cube within a mesh of larger cubes
	Faces could even come from different meshes or be contrived.
	'''
	# comment only
	# file.write('// brush "%s", "%s"\n' % (ob.name, ob.getData(name_only=1)))

	if PREF_GRID_SNAP.val:		format_vec= '( %d %d %d ) '
	else:						format_vec= '( %.8f %.8f %.8f ) '

	# unfortunately, O cannot simply make a 12 sided hexagon
	# I need to
	# - find the barycenter (mid point)
	# - make 12 tetrahedra
	# WOW Damn
	#

	mid = [0.0,0.0,0.0]
	numV=0
	numF=0
	for f in faces:
		print "Face %03d"%(numF)
		for v in f.v[2::-1]:
			print v
			mid[0] += v.co[0]
			mid[1] += v.co[1]
			mid[2] += v.co[2]
			numV+=1
		numF+=1

	mid[0] /= (float(numV))
	mid[1] /= (float(numV))
	mid[2] /= (float(numV))
	print "Midpoint for tetrahedra: "+repr(mid)

	# write for each face a tetrahedron
	for f in faces:
		file.write('// map tetrahedron brush from triangulated cube\n{\n')
		# from 4 verts this gets them in reversed order and only 3 of them
		# 0,1,2,3 -> 2,1,0
		#
		# 2011-02-01 OK: this was previously so
		verts = []
		for v in f.v[2::-1]:
			file.write(format_vec % tuple(v.co) )
			verts.append(tuple(v.co))

		try:	mode= f.mode
		except:	mode= 0



		if mode & Mesh.FaceModes.INVISIBLE:
			file.write(PREF_INVIS_TEX.val)
		else:
			try:	image= f.image
			except:	image= None

			if image:	file.write(sys.splitext(sys.basename(image.filename))[0])
			else:		file.write(PREF_TERR_TEX.val)

		# Texture stuff ignored for now
		file.write(PREF_DEF_TEX_OPTS.val)


		# now we need to write the three other sides of the tetrahedron
		# tetrahedron 2
		file.write(format_vec % verts[0])
		file.write(format_vec % tuple(mid))
		file.write(format_vec % verts[1])
		if mode & Mesh.FaceModes.INVISIBLE:
			file.write(PREF_INVIS_TEX.val)
		else:
			try:	image= f.image
			except:	image= None

			if image:	file.write(sys.splitext(sys.basename(image.filename))[0])
			else:		file.write(PREF_TERR_TEX.val)
		# Texture stuff ignored for now
		file.write(PREF_DEF_TEX_OPTS.val)
		#
		# tetrahedron 3
		file.write(format_vec % verts[1])
		file.write(format_vec % tuple(mid))
		file.write(format_vec % verts[2])
		if mode & Mesh.FaceModes.INVISIBLE:
			file.write(PREF_INVIS_TEX.val)
		else:
			try:	image= f.image
			except:	image= None

			if image:	file.write(sys.splitext(sys.basename(image.filename))[0])
			else:		file.write(PREF_TERR_TEX.val)
		# Texture stuff ignored for now
		file.write(PREF_DEF_TEX_OPTS.val)
		#
		# tetrajedron 4
		file.write(format_vec % verts[2])
		file.write(format_vec % tuple(mid))
		file.write(format_vec % verts[0])
		if mode & Mesh.FaceModes.INVISIBLE:
			file.write(PREF_INVIS_TEX.val)
		else:
			try:	image= f.image
			except:	image= None

			if image:	file.write(sys.splitext(sys.basename(image.filename))[0])
			else:		file.write(PREF_TERR_TEX.val)

		# Texture stuff ignored for now
		file.write(PREF_DEF_TEX_OPTS.val)

		#
		#
		#
		# end the tetrahedron brush
		file.write('}\n')



def round_vec(v):
	if PREF_GRID_SNAP.val:
		return round(v.x), round(v.y), round(v.z)
	else:
		return tuple(v)

def write_face2brush(file, faces):
	'''
	takes a face and writes it as a brush
	each face is a cube/brush


	OK: but different now:
	right now, each face is made into a prism (?)
	'''

	# translationsnormals? use PREF_PREJ_[X;Y;Z]
	trN = [PREF_PROJ_X.val, PREF_PROJ_Y.val, PREF_PROJ_Z.val]

	# schwerpunkt
	mid = [0.0,0.0,0.0]
	numV=0
	numF=0
	for f in faces:
		#print "Face %03d"%(numF)
		for v in f.v[2::-1]:
			#print v
			mid[0] += v.co[0]
			mid[1] += v.co[1]
			mid[2] += v.co[2]
			numV+=1
		numF+=1

	# new verts that give the face a thickness
	dist= PREF_SCALE.val * PREF_FACE_THICK.val
	#print "Translation Dist: "+repr(dist)
	#print "TrN: "+repr(trN)
	for i in range(0,3):
		trN[i] *= dist
	#print "TrN scaled: "+repr(trN)

	# das alte expoerter hat das per face gemacht, wir machen das auch, na klar
	for face in faces:

		if PREF_GRID_SNAP.val:		format_vec= '( %4d %4d %4d ) '
		else:						format_vec= '( %4.8f %4.8f %4.8f ) '

		try:	mode= face.mode
		except:	mode= 0


		# original verts as tuples for writing
		orig_vco= [tuple(v.co) for v in face]


		#new_vco= [round_vec(v.co - (tuple(trN) * dist)) for v in face]
		new_vco = []
		for ve in orig_vco:

			#print "Old: "+repr(ve)
			#for zur in range(0,3):
			#	print ve[zur]

			vNew= [ve[0],ve[1],ve[2]]
			for vi in range(0,3):
				if trN[vi] != 0: vNew[vi] = trN[vi]
			new_vco.append(tuple(vNew) )
			#print "New: "+repr(new_vco[-1])
		#print orig_vco
		#print new_vco



		file.write('// prism brush from face\n{\n')

		# bottom
		#for co in new_vco[:3]:
		#	file.write(format_vec % co )
		file.write(format_vec % new_vco[0] )
		file.write(format_vec % new_vco[1] )
		file.write(format_vec % new_vco[2] )
		image_text= PREF_INVIS_TEX.val
		file.write(image_text)
		file.write(PREF_DEF_TEX_OPTS.val)


		# front
		for co in orig_vco[2::-1]:
			file.write(format_vec % co )
		image_text= PREF_TERR_TEX.val
		file.write(image_text)
		file.write(PREF_DEF_TERR_OPTS.val)



		# sides.
		if len(orig_vco)==3: # Tri, it seemms tri brushes are supported.
			index_pairs= ((0,1), (1,2), (2,0))
		else: # cubes should also work
			index_pairs= ((0,1), (1,2), (2,3), (3,0))

		for i1, i2 in index_pairs:
			for co in orig_vco[i1], orig_vco[i2], new_vco[i2]:
				file.write( format_vec %  co )
			image_text= PREF_INVIS_TEX.val
			file.write(image_text)
			file.write(PREF_DEF_TEX_OPTS.val)

		file.write('}\n')

def is_cube_facegroup(faces):
	'''
	Returens a bool, true if the faces make up a cube
	'''
	# CHANGED
	# cube must have 12 faces (since 6 sides a two triangles)

	print "CHECK for 12sided cube goodness"

	if len(faces) != 12:
		print "Did not find 12 faces, found %d"%(len(faces))
		return False

	# Check for quads and that there are 6 unique verts
	verts= {}
	for f in faces:
		if len(f)!= 3: # must be triangles
			print "Found face with length: %d"%(len(f))
			return False

		for v in f:
			verts[v.index]= 0

	if len(verts) != 8:
		print "Did not find 8 vertices: Found %d"%(len(verts))
		return False

	# Now check that each vert has 3 face users
	for f in faces:
		for v in f:
			verts[v.index] += 1
	'''
	for v in verts.itervalues():
		if v != 3: # vert has 3 users?
			return False
	'''

	# Could we check for 12 unique edges??, probably not needed.
	return True

def is_tricyl_facegroup(faces):
	'''
	is the face group a tri cylinder
	Returens a bool, true if the faces make an extruded tri solid
	'''

	# cube must have 5 faces
	if len(faces) != 5:
		print '1'
		return False

	# Check for quads and that there are 6 unique verts
	verts= {}
	tottri= 0
	for f in faces:
		if len(f)== 3:
			tottri+=1

		for v in f:
			verts[v.index]= 0

	if len(verts) != 6 or tottri != 2:
		return False

	# Now check that each vert has 3 face users
	for f in faces:
		for v in f:
			verts[v.index] += 1

	for v in verts.itervalues():
		if v != 3: # vert has 3 users?
			return False

	# Could we check for 12 unique edges??, probably not needed.
	return True

def write_node_map(file, ob):
	'''
	Writes the properties of an object (empty in this case)
	as a MAP node as long as it has the property name - classname
	returns True/False based on weather a node was written
	'''
	props= [(p.name, p.data) for p in ob.game_properties]

	IS_MAP_NODE= False
	for name, value in props:
		if name=='classname':
			IS_MAP_NODE= True
			break

	if not IS_MAP_NODE:
		return False

	# Write a node
	file.write('{\n')
	for name_value in props:
		file.write('"%s" "%s"\n' % name_value)
	if PREF_GRID_SNAP.val:
		file.write('"origin" "%d %d %d"\n' % tuple([round(axis*PREF_SCALE.val) for axis in ob.getLocation('worldspace')]) )
	else:
		file.write('"origin" "%.6f %.6f %.6f"\n' % tuple([axis*PREF_SCALE.val for axis in ob.getLocation('worldspace')]) )
	file.write('}\n')
	return True


def export_map(filepath):

	pup_block = [\
	('Scale:', PREF_SCALE, 1, 1000, 'Scale the blender scene by this value.'),\
	('Face Width:', PREF_FACE_THICK, 4, 1024, 'Thickness of faces exported as brushes.'),\
	('Project Dir X:', PREF_PROJ_X, -1.0, 1.0,'Project Normal X'),\
	('Project Dir Y:', PREF_PROJ_Y, -1.0, 1.0,'Project Normal Y'),\
	('Project Dir Z:', PREF_PROJ_Z, -1.0, 1.0,'Project Normal Z'),\
	('Grid Snap', PREF_GRID_SNAP, 'snaps floating point values to whole numbers.'),\
	'Null Texture',\
	('', PREF_TERR_TEX, 1, 128, 'Export terrain faces with this texture'),\
	'Unseen Texture',\
	('', PREF_INVIS_TEX, 1, 128, 'Export invisible faces with this texture'),\
	]

	if not Draw.PupBlock('map export', pup_block):
		return

	Window.WaitCursor(1)
	time= sys.time()
	print 'Triangulated Blender Cubes UFORadiantMap Exporter 0.1'
	file= open(filepath, 'w')


	obs_mesh= []
	obs_lamp= []
	obs_surf= []
	obs_empty= []

	SCALE_MAT= Mathutils.Matrix()
	SCALE_MAT[0][0]= SCALE_MAT[1][1]= SCALE_MAT[2][2]= PREF_SCALE.val

	dummy_mesh= Mesh.New()

	TOTBRUSH= TOTLAMP= TOTNODE= 0

	for ob in Object.GetSelected():
		type= ob.type
		if type == 'Mesh':		obs_mesh.append(ob)
		#elif type == 'Surf':	obs_surf.append(ob)
		elif type == 'Lamp':	obs_lamp.append(ob)
		#elif type == 'Empty':	obs_empty.append(ob)

	if obs_mesh or obs_surf:
		# brushes and surf's must be under worldspan
		file.write('\n// entity 0\n')
		file.write('{\n')
		file.write('"classname" "worldspawn"\n')


	print '\twriting cubes from meshes'
	for ob in obs_mesh:
		dummy_mesh.getFromObject(ob.name)

		#print len(mesh_split2connected(dummy_mesh))

		# Is the object 1 cube? - object-is-a-brush
		dummy_mesh.transform(ob.matrixWorld*SCALE_MAT) # 1 to tx the normals also

		if PREF_GRID_SNAP.val:
			for v in dummy_mesh.verts:
				co= v.co
				co.x= round(co.x)
				co.y= round(co.y)
				co.z= round(co.z)

		# High quality normals
		BPyMesh.meshCalcNormals(dummy_mesh)

		# Split mesh into connected regions
		for face_group in BPyMesh.mesh2linkedFaces(dummy_mesh):
			if is_cube_facegroup(face_group):
				write_cube2brush(file, face_group)
				TOTBRUSH+=1
			else:
				write_face2brush(file, face_group)

			print 'warning, not exporting "%s" it is not a cube' % ob.name


	dummy_mesh.verts= None


	valid_dims= 3,5,7,9,11,13,15
	for ob in obs_surf:
		'''
		Surf, patches
		'''
		surf_name= ob.getData(name_only=1)
		data= Curve.Get(surf_name)
		mat = ob.matrixWorld*SCALE_MAT

		# This is what a valid patch looks like

		"""
// brush 0
{
patchDef2
{
NULL
( 3 3 0 0 0 )
(
( ( -64 -64 0 0 0 ) ( -64 0 0 0 -2 ) ( -64 64 0 0 -4 ) )
( ( 0 -64 0 2 0 ) ( 0 0 0 2 -2 ) ( 0 64 0 2 -4 ) )
( ( 64 -64 0 4 0 ) ( 64 0 0 4 -2 ) ( 80 88 0 4 -4 ) )
)
}
}
		"""
		for i, nurb in enumerate(data):
			u= nurb.pointsU
			v= nurb.pointsV
			if u in valid_dims and v in valid_dims:

				file.write('// brush %d surf_name\n' % i)
				file.write('{\n')
				file.write('patchDef2\n')
				file.write('{\n')
				file.write('NULL\n')
				file.write('( %d %d 0 0 0 )\n' % (u, v) )
				file.write('(\n')

				u_iter = 0
				for p in nurb:

					if u_iter == 0:
						file.write('(')

					u_iter += 1

					# add nmapping 0 0 ?
					if PREF_GRID_SNAP.val:
						file.write(' ( %d %d %d 0 0 )' % round_vec(Mathutils.Vector(p[0:3]) * mat))
					else:
						file.write(' ( %.6f %.6f %.6f 0 0 )' % tuple(Mathutils.Vector(p[0:3]) * mat))

					# Move to next line
					if u_iter == u:
						file.write(' )\n')
						u_iter = 0

				file.write(')\n')
				file.write('}\n')
				file.write('}\n')


				# Debugging
				# for p in nurb: print 'patch', p

			else:
				print "NOT EXPORTING PATCH", surf_name, u,v, 'Unsupported'


	if obs_mesh or obs_surf:
		file.write('}\n') # end worldspan


	print '\twriting lamps'
	for ob in obs_lamp:
		print '\t\t%s' % ob.name
		lamp= ob.data
		file.write('{\n')
		file.write('"classname" "light"\n')
		file.write('"light" "%.6f"\n' % (lamp.dist* PREF_SCALE.val))
		if PREF_GRID_SNAP.val:
			file.write('"origin" "%d %d %d"\n' % tuple([round(axis*PREF_SCALE.val) for axis in ob.getLocation('worldspace')]) )
		else:
			file.write('"origin" "%.6f %.6f %.6f"\n' % tuple([axis*PREF_SCALE.val for axis in ob.getLocation('worldspace')]) )
		file.write('"_color" "%.6f %.6f %.6f"\n' % tuple(lamp.col))
		file.write('"style" "0"\n')
		file.write('}\n')
		TOTLAMP+=1


	print '\twriting empty objects as nodes'
	for ob in obs_empty:
		if write_node_map(file, ob):
			print '\t\t%s' % ob.name
			TOTNODE+=1
		else:
			print '\t\tignoring %s' % ob.name

	Window.WaitCursor(0)

	print 'Exported Map in %.4fsec' % (sys.time()-time)
	print 'Brushes: %d  Nodes: %d  Lamps %d\n' % (TOTBRUSH, TOTNODE, TOTLAMP)


def main():
	Window.FileSelector(export_map, 'EXPORT MAP', '*.map')

if __name__ == '__main__': main()
# export_map('/foo.map')



# Test script to compare transform-matrix vlaues in Blender
# Load in blender text editor, select one or more empties and hit [ALT][P]

# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****

import Blender
from Blender import *
from Blender.Window import *
from Blender.Object import *
from Blender.Mathutils import *

import math
import struct, string
from types import *

objects = Blender.Object.GetSelected()
for object in objects:
	if object.getType()=="Empty":
		origin = object.getLocation('worldspace')
		matrix = object.getMatrix('worldspace')
		print "Name", object.name
		print " Origin:", origin[0], " ",origin[1]," ",origin[2]
		print " Matrix:"
		print "", matrix[0][0], " ",matrix[0][1]," ",matrix[0][2]
		print "", matrix[1][0], " ",matrix[1][1]," ",matrix[1][2]
		print "", matrix[2][0], " ",matrix[2][1]," ",matrix[2][2]
